# Copyright (c) 2010-2024, Manfred Moitzi
# License: MIT License
import pytest

import os
import math
from ezdxf.addons.acadctb import *


class TestPlotStyle:
    def test_init(self):
        style = PlotStyle(
            0,
            dict(
                description="memo",
                color_policy=3,
                physical_pen_number=7,
                virtual_pen_number=8,
                screen=99,
                linepattern_size=0.7,
                linetype=17,
                adaptive_linetype=False,
                lineweight=19,
                end_style=1,
                join_style=2,
                fill_style=3,
            ),
        )
        assert style.description == "memo"
        assert style.name == "Color_1"
        assert style.localized_name == "Color_1"
        assert style.dithering is True
        assert style.grayscale is True
        assert style.physical_pen_number == 7
        assert style.virtual_pen_number == 8
        assert style.screen == 99
        assert style.linepattern_size == 0.7
        assert style.adaptive_linetype is False
        assert style.lineweight == 19
        assert style.end_style == 1
        assert style.join_style == 2
        assert style.fill_style == 3

    def test_get_set_color(self):
        style = PlotStyle(0)
        style.color = (123, 23, 77)
        assert style.color == (123, 23, 77)

    def test_object_color(self):
        style = PlotStyle(0)
        style.color = (123, 77, 1)
        style.set_object_color()
        assert style.has_object_color()

    def test_dithering(self):
        style = PlotStyle(0)
        style.dithering = True
        assert style.dithering
        style.dithering = False
        assert style.dithering is False

    def test_grayscale(self):
        style = PlotStyle(0)
        style.grayscale = True
        assert style.grayscale
        style.grayscale = False
        assert style.grayscale is False

    def test_dxf_color_index(self):
        style = PlotStyle(0)
        assert style.aci == 1

    def test_set_lineweight(self):
        styles = ColorDependentPlotStyles()
        style = styles[5]
        style.set_lineweight(0.5)
        assert style.lineweight == 13

    def test_write(self):
        expected = (
            " 0{\n"
            '  name="Color_1\n'
            '  localized_name="Color_1\n'
            '  description="\n'
            "  color=-1\n"
            "  color_policy=1\n"
            "  physical_pen_number=0\n"
            "  virtual_pen_number=0\n"
            "  screen=100\n"
            "  linepattern_size=0.5\n"
            "  linetype=31\n"
            "  adaptive_linetype=TRUE\n"
            "  lineweight=0\n"
            "  fill_style=73\n"
            "  end_style=4\n"
            "  join_style=5\n"
            " }\n"
        )
        style = PlotStyle(0)
        fp = StringIO()
        style.write(fp)
        result = fp.getvalue()
        fp.close()
        assert result == expected


class TestColorDependentPlotStyles:
    @pytest.fixture
    def styles(self):
        return ColorDependentPlotStyles(description="TestCase")

    def test_set_style(self, styles):
        style = styles.new_style(3, dict(description="TestCase"))
        assert style.named_color is False
        assert style.color_type is None
        assert style.description == "TestCase"

    def test_get_style(self, styles):
        styles.new_style(3, dict(description="TestCase"))
        assert styles[3].description == "TestCase"

    def test_get_color(self, styles):
        style = styles.new_style(4, dict(description="TestCase"))
        style.color = (123, 17, 99)
        assert style.named_color is False
        assert style.color_type is COLOR_RGB
        assert styles[4].color == (123, 17, 99)

    def test_get_lineweight(self, styles):
        style = styles.new_style(5, dict(description="TestCase"))
        style.set_lineweight(0.70)
        assert math.isclose(styles.get_lineweight(5), 0.70, abs_tol=1e-6)

    def test_get_lineweight_none(self, styles):
        style = styles.new_style(5, dict(description="TestCase"))
        style.set_lineweight(0.0)
        assert styles.get_lineweight(5) is None

    def test_get_lineweight_index(self, styles):
        assert styles.get_lineweight_index(0.50) == 13

    def test_get_table_lineweight(self, styles):
        assert styles.get_table_lineweight(13) == 0.50

    def test_set_table_lineweight(self, styles):
        styles.set_table_lineweight(7, 1.70)
        assert math.isclose(styles.get_table_lineweight(7), 1.70, abs_tol=1e-6)

    def test_set_table_lineweight_index_error(self, styles):
        index_out_of_range = len(styles.lineweights) + 7
        index = styles.set_table_lineweight(index_out_of_range, 7.77)
        assert math.isclose(styles.get_table_lineweight(index), 7.77, abs_tol=1e-6)

    def test_write_header(self):
        expected = (
            'description="\n'
            "aci_table_available=TRUE\n"
            "scale_factor=1.0\n"
            "apply_factor=FALSE\n"
            "custom_lineweight_display_units=0\n"
        )
        styles = ColorDependentPlotStyles()
        fp = StringIO()
        styles._write_header(fp)
        result = fp.getvalue()
        fp.close()
        assert result == expected

    def test_write_aci_table(self):
        expected = (
            "aci_table{\n"
            + "\n".join((' %s="Color_%d' % (index, index + 1) for index in range(255)))
            + "\n}\n"
        )
        styles = ColorDependentPlotStyles()
        fp = StringIO()
        styles._write_aci_table(fp)
        result = fp.getvalue()
        fp.close()
        assert str(result) == str(expected)

    def test_write_lineweights(self):
        expected = (
            "custom_lineweight_table{\n"
            " 0=0.00\n 1=0.05\n 2=0.09\n 3=0.10\n 4=0.13\n"
            " 5=0.15\n 6=0.18\n 7=0.20\n 8=0.25\n 9=0.30\n"
            " 10=0.35\n 11=0.40\n 12=0.45\n 13=0.50\n 14=0.53\n"
            " 15=0.60\n 16=0.65\n 17=0.70\n 18=0.80\n 19=0.90\n"
            " 20=1.00\n 21=1.06\n 22=1.20\n 23=1.40\n 24=1.58\n"
            " 25=2.00\n 26=2.11\n}\n"
        )
        styles = ColorDependentPlotStyles()
        fp = StringIO()
        styles._write_lineweights(fp)
        result = fp.getvalue()
        fp.close()
        assert str(result) == str(expected)


class TestNamedPlotStyles:
    @pytest.fixture
    def stb(self):
        return NamedPlotStyles(description="TestCase")

    def test_set_style(self, stb):
        style = stb.new_style("ezdxf", dict(description="TestCase"))
        assert style.named_color is True
        assert style.color_type is None
        assert style.description == "TestCase"

    def test_get_style(self, stb):
        stb.new_style("mozman", dict(description="ezdxf comment"))
        assert stb["mozman"].description == "ezdxf comment"

    def test_get_color(self, stb):
        style = stb.new_style("mozman2")
        style.color = (123, 17, 99)
        assert style.named_color is True
        # named colors has always color type: COLOR_ACI, why?
        assert style.color_type == COLOR_ACI
        assert stb["mozman2"].color == (123, 17, 99)

    def test_get_lineweight(self, stb):
        style = stb.new_style("mozman5")
        style.set_lineweight(0.70)
        assert math.isclose(stb.get_lineweight("mozman5"), 0.70, abs_tol=1e-6)

    def test_get_lineweight_none(self, stb):
        style = stb.new_style("mozman4")
        style.set_lineweight(0.0)
        assert stb.get_lineweight("mozman4") is None

    def test_get_lineweight_index(self, stb):
        assert stb.get_lineweight_index(0.50) == 13

    def test_get_table_lineweight(self, stb):
        assert stb.get_table_lineweight(13) == 0.50

    def test_set_table_lineweight(self, stb):
        stb.set_table_lineweight(7, 1.70)
        assert math.isclose(stb.get_table_lineweight(7), 1.70, abs_tol=1e-6)

    def test_set_table_lineweight_index_error(self, stb):
        index_out_of_range = len(stb.lineweights) + 7
        index = stb.set_table_lineweight(index_out_of_range, 7.77)
        assert math.isclose(stb.get_table_lineweight(index), 7.77, abs_tol=1e-6)

    def test_write_header(self):
        expected = (
            'description="\n'
            "aci_table_available=FALSE\n"
            "scale_factor=1.0\n"
            "apply_factor=FALSE\n"
            "custom_lineweight_display_units=0\n"
        )
        stb = NamedPlotStyles()
        fp = StringIO()
        stb._write_header(fp)
        result = fp.getvalue()
        fp.close()
        assert result == expected

    def test_write_lineweights(self):
        expected = (
            "custom_lineweight_table{\n"
            " 0=0.00\n 1=0.05\n 2=0.09\n 3=0.10\n 4=0.13\n"
            " 5=0.15\n 6=0.18\n 7=0.20\n 8=0.25\n 9=0.30\n"
            " 10=0.35\n 11=0.40\n 12=0.45\n 13=0.50\n 14=0.53\n"
            " 15=0.60\n 16=0.65\n 17=0.70\n 18=0.80\n 19=0.90\n"
            " 20=1.00\n 21=1.06\n 22=1.20\n 23=1.40\n 24=1.58\n"
            " 25=2.00\n 26=2.11\n}\n"
        )
        stb = NamedPlotStyles()
        fp = StringIO()
        stb._write_lineweights(fp)
        result = fp.getvalue()
        fp.close()
        assert str(result) == str(expected)


class TestCTBImport:
    @pytest.fixture(scope="class")
    def ctb(self):
        path, name = os.path.split(__file__)
        ctbfile = os.path.join(path, "ctbtest.ctb")
        return load(ctbfile)

    def test_ctb_attribs(self, ctb):
        styles = ctb
        assert styles.description == "Comment"
        assert styles.apply_factor is True
        assert styles.scale_factor == 2

    def test_lineweight_table(self, ctb):
        lineweights = ctb.lineweights
        for default_lw, ctb_lw in zip(DEFAULT_LINE_WEIGHTS, lineweights):
            assert math.isclose(default_lw, ctb_lw, abs_tol=1e-6)

    def test_style_1(self, ctb):
        """all attribs are user defined."""
        style = ctb[1]
        assert isinstance(style, PlotStyle)
        assert style.aci == 1
        assert style.color_type == COLOR_RGB
        assert style.color == (235, 135, 20)
        assert style.dithering is True
        assert style.grayscale is True
        assert style.has_object_color() is False
        assert style.physical_pen_number is 11
        assert style.virtual_pen_number == 5
        assert style.screen == 95
        assert style.linetype == 1
        assert style.end_style == END_STYLE_SQUARE
        assert style.join_style == JOIN_STYLE_ROUND
        assert style.fill_style == FILL_STYLE_SOLID

    def test_style_3(self, ctb):
        """all attribs are default or auto"""
        style = ctb[3]
        assert isinstance(style, PlotStyle)
        assert style.aci == 3
        assert style.color_type is None
        assert style.color is None
        assert style.dithering is True
        assert style.grayscale is False
        assert style.has_object_color() is True
        assert style.physical_pen_number == AUTOMATIC
        assert style.virtual_pen_number == AUTOMATIC
        assert style.screen == 100
        assert style.linetype == OBJECT_LINETYPE
        assert style.end_style == END_STYLE_OBJECT
        assert style.join_style == JOIN_STYLE_OBJECT
        assert style.fill_style == FILL_STYLE_OBJECT


class TestCTBExport:
    def test_create_ctb(self, tmpdir):
        filename = str(tmpdir.join("new.ctb"))
        styles = ColorDependentPlotStyles("TestCTB")
        styles.save(filename)
        assert os.path.exists(filename)


class TestSTBImport:
    @pytest.fixture(scope="class")
    def stb(self):
        path, name = os.path.split(__file__)
        stb_file = os.path.join(path, "stbtest.stb")
        return load(stb_file)

    def test_stb_attribs(self, stb):
        assert stb.description == "STB file for ezdxf testing"
        assert stb.apply_factor is True
        assert stb.scale_factor == 2

    def test_normal(self, stb):
        """All attributes of 'Normal' are fixed default values."""
        style = stb["Normal"]
        assert isinstance(style, PlotStyle)
        assert style.index == 0
        assert style.color is None
        assert style.dithering is True
        assert style.grayscale is False
        assert style.has_object_color() is True
        assert style.color_type is None
        assert style.physical_pen_number == AUTOMATIC
        assert style.virtual_pen_number == AUTOMATIC
        assert style.screen == 100
        assert style.linetype == OBJECT_LINETYPE
        assert style.lineweight == OBJECT_LINEWEIGHT
        assert style.end_style == END_STYLE_OBJECT
        assert style.join_style == JOIN_STYLE_OBJECT
        assert style.fill_style == FILL_STYLE_OBJECT

    def test_style_1(self, stb):
        """All attribs are user defined."""
        style = stb["Style_1"]
        assert isinstance(style, PlotStyle)
        assert style.name == "Style_1"
        assert style.localized_name == "Style 1"
        assert style.color_type == COLOR_ACI  # ???
        assert style.color == (235, 135, 20)
        assert style.named_color is False  # why?
        assert style.dithering is True
        assert style.grayscale is True
        assert style.has_object_color() is False
        assert style.physical_pen_number == 11
        assert style.virtual_pen_number == 5
        assert style.screen == 95
        assert style.linetype == 1
        assert style.end_style == END_STYLE_SQUARE
        assert style.join_style == JOIN_STYLE_ROUND
        assert style.fill_style == FILL_STYLE_SOLID

    def test_mozman(self, stb):
        """All attribs are user defined."""
        style = stb["mozman"]
        assert isinstance(style, PlotStyle)
        assert style.name == "mozman"
        assert style.localized_name == "mozman"
        assert style.color_type == COLOR_ACI
        assert style.color == (255, 0, 255)
        assert style.named_color is True
        assert style.dithering is True
        assert style.grayscale is False
        assert style.has_object_color() is False
        assert style.physical_pen_number == AUTOMATIC
        assert style.virtual_pen_number == AUTOMATIC
        assert style.screen == 90
        assert style.linetype == 4
        assert style.end_style == END_STYLE_SQUARE
        assert style.join_style == JOIN_STYLE_MITER
        assert style.fill_style == FILL_STYLE_CHECKERBOARD


class TestSTBExport:
    def test_create_ctb(self, tmpdir):
        filename = str(tmpdir.join("new.stb"))
        styles = NamedPlotStyles("TestSTB")
        styles.save(filename)
        assert os.path.exists(filename)


class TestFunctions:
    def test_color_name(self):
        assert color_name(0) == "Color_1"

    def test_get_bool(self):
        assert get_bool(True)
        assert get_bool("true")
        assert get_bool(False) is False
        assert get_bool("false") is False
        pytest.raises(ValueError, get_bool, "falsch")


def test_parser_ignores_unknown_appendix():
    parser = PlotStyleFileParser(
        r"""custom_lineweight_table{
 0=0.0076200000000 (+7.Z+"8V?S_LC )
 1=0.0254
 6=0.17780000000   (  44F=@";"QC\\SM0)
 7=0.2
 10=0.35560000000 (44F=@";"UC_SN )
 11=0.4
}
"""
    )
    table = parser.get("custom_lineweight_table", {})
    assert table["0"] == "0.0076200000000"
    assert table["1"] == "0.0254"
    assert table["6"] == "0.17780000000"
    assert table["7"] == "0.2"
    assert table["10"] == "0.35560000000"
    assert table["11"] == "0.4"


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2018-2012, Manfred Moitzi
# License: MIT License
import pytest
import os
import ezdxf
from ezdxf.lldxf.const import versions_supported_by_new


@pytest.fixture(params=versions_supported_by_new)
def drawing(request):
    return ezdxf.new(request.param)


def add_line_entities(layout, offset):
    for color in range(1, 256):
        layout.add_line(
            (offset + 0, color),
            (offset + 50, color),
            {"color": color, "layer": "Träger"},
        )


def test_basic_graphics(drawing, tmpdir):
    add_line_entities(drawing.modelspace(), 0)
    add_line_entities(drawing.layout(), 70)
    filename = str(tmpdir.join("basic_graphics_%s.dxf" % drawing.dxfversion))
    try:
        drawing.saveas(filename)
    except ezdxf.DXFError as e:
        pytest.fail(
            "DXFError: {0} for DXF version {1}".format(
                str(e), drawing.dxfversion
            )
        )
    assert os.path.exists(filename) is True



# Copyright 2018-2021, Manfred Moitzi
# License: MIT License
import pytest
import ezdxf

DXF_PATH = ezdxf.options.test_files_path / "DXF_with_ACIS_data"
SURFACES = DXF_PATH / "All_Surfaces_R2010.dxf"
DELIMITER = "\n" + "=" * 80 + "\n"


@pytest.mark.skipif(not SURFACES.exists(), reason=f"File {SURFACES} not found")
def test_get_acis_data_from_surfaces():
    dwg = ezdxf.readfile(SURFACES)
    msp = dwg.modelspace()

    with open(DXF_PATH / "All_Surfaces_R2010.sat", "wt") as f:
        for surface in msp.query(
            "SURFACE SWEPTSURFACE REVOLVEDSURFACE LOFTEDSURFACE EXTRUDEDSURFACE"
        ):
            f.write(DELIMITER)
            f.write(str(surface) + ":")
            f.write(DELIMITER)
            f.write(surface.tostring())



# Created: 2019-02-15
# Copyright (c) 2019-2020, Manfred Moitzi
# License: MIT License
import pytest
from ezdxf.entitydb import EntitySpace


class Entity:
    def __init__(self, value):
        self.value = value
        self.is_alive = True


@pytest.fixture
def space():
    return EntitySpace(Entity(p) for p in [1, 4, 5, 6, 76, -4, 7])


def test_init(space):
    assert len(EntitySpace()) == 0
    assert len(space) == 7


def test_existence(space):
    e = space[3]
    assert e in space

    assert len(space) == 7
    e.is_alive = False
    assert e not in space
    assert len(space) == 7, "still 7 items"

    space.purge()
    assert len(space) == 6, "removed dead entities"

    e = Entity(1)
    assert e not in space


def test_remove(space):
    e = space[3]
    space.remove(e)
    assert e not in space

    space.clear()
    assert len(space) == 0



# cython: language_level=3
# Copyright (c) 2020-2024, Manfred Moitzi
# License: MIT License
from typing import Sequence, Iterable, Tuple, TYPE_CHECKING, Iterator
from itertools import chain
import math
import numpy as np
import cython

from .vector cimport (
Vec2, Vec3, v3_normalize, v3_isclose, v3_cross, v3_dot,
)
from .vector import X_AXIS, Y_AXIS, Z_AXIS, NULLVEC

from libc.math cimport fabs, sin, cos, tan

if TYPE_CHECKING:
    from ezdxf.math import UVec

cdef extern from "constants.h":
    const double ABS_TOL
    const double REL_TOL

cdef double[16] IDENTITY = [
    1.0, 0.0, 0.0, 0.0,
    0.0, 1.0, 0.0, 0.0,
    0.0, 0.0, 1.0, 0.0,
    0.0, 0.0, 0.0, 1.0
]

cdef void set_floats(double *m, object values) except *:
    cdef int i = 0
    for v in values:
        if i < 16:  # Do not write beyond array bounds
            m[i] = v
        i += 1
    if i != 16:
        raise ValueError("invalid argument count")

cdef class Matrix44:
    def __cinit__(self, *args):
        cdef int nargs = len(args)
        if nargs == 0:  # default constructor Matrix44(): fastest setup
            self.m = IDENTITY  # memcopy!
        elif nargs == 1:  # 16 numbers: slow setup
            set_floats(self.m, args[0])
        elif nargs == 4:  # 4 rows of 4 numbers: slowest setup
            set_floats(self.m, chain(*args))
        else:
            raise ValueError("invalid argument count: 4 row vectors or "
                             "iterable of 16 numbers")

    def __reduce__(self):
        return Matrix44, (tuple(self),)

    def __getitem__(self, tuple index) -> float:
        cdef int row = index[0]
        cdef int col = index[1]
        cdef int i = row * 4 + col

        if 0 <= i < 16 and 0 <= col < 4:
            return self.m[i]
        else:
            raise IndexError(f'index out of range: {index}')

    def __setitem__(self, tuple index, double value):
        cdef int row = index[0]
        cdef int col = index[1]
        cdef int i = row * 4 + col

        if 0 <= i < 16 and 0 <= col < 4:
            self.m[i] = value
        else:
            raise IndexError(f'index out of range: {index}')

    def __iter__(self):
        cdef int i
        for i in range(16):
            yield self.m[i]

    def __repr__(self) -> str:
        def format_row(row):
            return "(%s)" % ", ".join(str(value) for value in row)

        return "Matrix44(%s)" % \
               ", ".join(format_row(row) for row in self.rows())

    def get_2d_transformation(self) -> Tuple[float, ...]:
        cdef double *m = self.m
        return m[0], m[1], 0.0, m[4], m[5], 0.0, m[12], m[13], 1.0

    @staticmethod
    def from_2d_transformation(components: Sequence[float]) -> Matrix44:
        if len(components) != 6:
            raise ValueError(
                "First 2 columns of a 3x3 matrix required: m11, m12, m21, m22, m31, m32"
            )

        m44 = Matrix44()
        m44.m[0] = components[0]
        m44.m[1] = components[1]
        m44.m[4] = components[2]
        m44.m[5] = components[3]
        m44.m[12] = components[4]
        m44.m[13] = components[5]
        return m44

    def get_row(self, int row) -> Tuple[float, ...]:
        cdef int index = row * 4
        if 0 <= index < 13:
            return self.m[index], self.m[index + 1], self.m[index + 2], self.m[
                index + 3]
        else:
            raise IndexError(f'invalid row index: {row}')

    def set_row(self, int row, values: Sequence[float]) -> None:
        cdef Py_ssize_t count = len(values)
        cdef Py_ssize_t start = row * 4
        cdef Py_ssize_t i
        if 0 <= row < 4:
            if count > 4:
                count = 4
            for i in range(count):
                self.m[start + i] = values[i]
        else:
            raise IndexError(f'invalid row index: {row}')

    def get_col(self, int col) -> Tuple[float, ...]:
        if 0 <= col < 4:
            return self.m[col], self.m[col + 4], \
                   self.m[col + 8], self.m[col + 12]
        else:
            raise IndexError(f'invalid col index: {col}')

    def set_col(self, int col, values: Sequence[float]):
        cdef Py_ssize_t count = len(values)
        cdef Py_ssize_t i
        if 0 <= col < 4:
            if count > 4:
                count = 4
            for i in range(count):
                self.m[col + i * 4] = values[i]
        else:
            raise IndexError(f'invalid col index: {col}')

    def rows(self) -> Iterator[Tuple[float, ...]]:
        return (self.get_row(index) for index in (0, 1, 2, 3))

    def columns(self) -> Iterator[Tuple[float, ...]]:
        return (self.get_col(index) for index in (0, 1, 2, 3))

    def copy(self) -> Matrix44:
        cdef Matrix44 _copy = Matrix44()
        _copy.m = self.m
        return _copy

    __copy__ = copy

    @property
    def origin(self) -> Vec3:
        cdef Vec3 v = Vec3()
        v.x = self.m[12]
        v.y = self.m[13]
        v.z = self.m[14]
        return v

    @origin.setter
    def origin(self, v: UVec) -> None:
        cdef Vec3 origin = Vec3(v)
        self.m[12] = origin.x
        self.m[13] = origin.y
        self.m[14] = origin.z

    @property
    def ux(self) -> Vec3:
        return self.get_ux()

    cdef Vec3 get_ux(self):
        cdef Vec3 v = Vec3()
        v.x = self.m[0]
        v.y = self.m[1]
        v.z = self.m[2]
        return v

    @property
    def uy(self) -> Vec3:
        return self.get_uy()

    cdef Vec3 get_uy(self):
        cdef Vec3 v = Vec3()
        v.x = self.m[4]
        v.y = self.m[5]
        v.z = self.m[6]
        return v

    @property
    def uz(self) -> Vec3:
        return self.get_uz()

    cdef Vec3 get_uz(self):
        cdef Vec3 v = Vec3()
        v.x = self.m[8]
        v.y = self.m[9]
        v.z = self.m[10]
        return v

    @property
    def is_cartesian(self) -> bool:
        cdef Vec3 x_axis = v3_cross(self.get_uy(), self.get_uz())
        return v3_isclose(x_axis, self.get_ux(), REL_TOL, ABS_TOL)

    @property
    def is_orthogonal(self) -> bool:
        cdef Vec3 ux = v3_normalize(self.get_ux(), 1.0)
        cdef Vec3 uy = v3_normalize(self.get_uy(), 1.0)
        cdef Vec3 uz = v3_normalize(self.get_uz(), 1.0)
        return fabs(v3_dot(ux, uy)) < 1e-9 and \
               fabs(v3_dot(ux, uz)) < 1e-9 and \
               fabs(v3_dot(uy, uz)) < 1e-9

    @staticmethod
    def scale(double sx, sy = None, sz = None) -> Matrix44:
        cdef Matrix44 mat = Matrix44()
        mat.m[0] = sx
        mat.m[5] = sx if sy is None else sy
        mat.m[10] = sx if sz is None else sz
        return mat

    @staticmethod
    def translate(double dx, double dy, double dz) -> Matrix44:
        cdef Matrix44 mat = Matrix44()
        mat.m[12] = dx
        mat.m[13] = dy
        mat.m[14] = dz
        return mat

    @staticmethod
    def x_rotate(double angle) -> Matrix44:
        cdef Matrix44 mat = Matrix44()
        cdef double cos_a = cos(angle)
        cdef double sin_a = sin(angle)
        mat.m[5] = cos_a
        mat.m[6] = sin_a
        mat.m[9] = -sin_a
        mat.m[10] = cos_a
        return mat

    @staticmethod
    def y_rotate(double angle) -> Matrix44:
        cdef Matrix44 mat = Matrix44()
        cdef double cos_a = cos(angle)
        cdef double sin_a = sin(angle)
        mat.m[0] = cos_a
        mat.m[2] = -sin_a
        mat.m[8] = sin_a
        mat.m[10] = cos_a
        return mat

    @staticmethod
    def z_rotate(double angle) -> Matrix44:
        cdef Matrix44 mat = Matrix44()
        cdef double cos_a = cos(angle)
        cdef double sin_a = sin(angle)
        mat.m[0] = cos_a
        mat.m[1] = sin_a
        mat.m[4] = -sin_a
        mat.m[5] = cos_a
        return mat

    @staticmethod
    def axis_rotate(axis: UVec, double angle) -> Matrix44:
        cdef Matrix44 mat = Matrix44()
        cdef double cos_a = cos(angle)
        cdef double sin_a = sin(angle)
        cdef double one_m_cos = 1.0 - cos_a
        cdef Vec3 _axis = Vec3(axis).normalize()
        cdef double x = _axis.x
        cdef double y = _axis.y
        cdef double z = _axis.z

        mat.m[0] = x * x * one_m_cos + cos_a
        mat.m[1] = y * x * one_m_cos + z * sin_a
        mat.m[2] = x * z * one_m_cos - y * sin_a

        mat.m[4] = x * y * one_m_cos - z * sin_a
        mat.m[5] = y * y * one_m_cos + cos_a
        mat.m[6] = y * z * one_m_cos + x * sin_a

        mat.m[8] = x * z * one_m_cos + y * sin_a
        mat.m[9] = y * z * one_m_cos - x * sin_a
        mat.m[10] = z * z * one_m_cos + cos_a

        return mat

    @staticmethod
    def xyz_rotate(double angle_x, double angle_y,
                   double angle_z) -> Matrix44:
        cdef Matrix44 mat = Matrix44()
        cdef double cx = cos(angle_x)
        cdef double sx = sin(angle_x)
        cdef double cy = cos(angle_y)
        cdef double sy = sin(angle_y)
        cdef double cz = cos(angle_z)
        cdef double sz = sin(angle_z)
        cdef double sxsy = sx * sy
        cdef double cxsy = cx * sy

        mat.m[0] = cy * cz
        mat.m[1] = sxsy * cz + cx * sz
        mat.m[2] = -cxsy * cz + sx * sz
        mat.m[4] = -cy * sz
        mat.m[5] = -sxsy * sz + cx * cz
        mat.m[6] = cxsy * sz + sx * cz
        mat.m[8] = sy
        mat.m[9] = -sx * cy
        mat.m[10] = cx * cy
        return mat

    @staticmethod
    def shear_xy(double angle_x = 0, double angle_y = 0) -> Matrix44:
        cdef Matrix44 mat = Matrix44()
        cdef double tx = tan(angle_x)
        cdef double ty = tan(angle_y)
        mat.m[1] = ty
        mat.m[4] = tx
        return mat

    @staticmethod
    def perspective_projection(double left, double right, double top,
                               double bottom, double near,
                               double far) -> Matrix44:
        cdef Matrix44 mat = Matrix44()
        mat.m[0] = (2. * near) / (right - left)
        mat.m[5] = (2. * near) / (top - bottom)
        mat.m[8] = (right + left) / (right - left)
        mat.m[9] = (top + bottom) / (top - bottom)
        mat.m[10] = -((far + near) / (far - near))
        mat.m[11] = -1
        mat.m[14] = -((2. * far * near) / (far - near))
        return mat

    @staticmethod
    def perspective_projection_fov(fov: float, aspect: float, near: float,
                                   far: float) -> Matrix44:
        vrange = near * math.tan(fov / 2.)
        left = -vrange * aspect
        right = vrange * aspect
        bottom = -vrange
        top = vrange
        return Matrix44.perspective_projection(left, right, bottom, top, near,
                                               far)

    @staticmethod
    def chain(*matrices: Matrix44) -> Matrix44:
        cdef Matrix44 transformation = Matrix44()
        for matrix in matrices:
            transformation *= matrix
        return transformation

    def __imul__(self, Matrix44 other) -> Matrix44:
        cdef double[16] m1 = self.m
        cdef double *m2 = other.m

        self.m[0] = m1[0] * m2[0] + m1[1] * m2[4] + m1[2] * m2[8] + \
                    m1[3] * m2[12]
        self.m[1] = m1[0] * m2[1] + m1[1] * m2[5] + m1[2] * m2[9] + \
                    m1[3] * m2[13]
        self.m[2] = m1[0] * m2[2] + m1[1] * m2[6] + m1[2] * m2[10] + \
                    m1[3] * m2[14]
        self.m[3] = m1[0] * m2[3] + m1[1] * m2[7] + m1[2] * m2[11] + \
                    m1[3] * m2[15]

        self.m[4] = m1[4] * m2[0] + m1[5] * m2[4] + m1[6] * m2[8] + \
                    m1[7] * m2[12]
        self.m[5] = m1[4] * m2[1] + m1[5] * m2[5] + m1[6] * m2[9] + \
                    m1[7] * m2[13]
        self.m[6] = m1[4] * m2[2] + m1[5] * m2[6] + m1[6] * m2[10] + \
                    m1[7] * m2[14]
        self.m[7] = m1[4] * m2[3] + m1[5] * m2[7] + m1[6] * m2[11] + \
                    m1[7] * m2[15]

        self.m[8] = m1[8] * m2[0] + m1[9] * m2[4] + m1[10] * m2[8] + \
                    m1[11] * m2[12]
        self.m[9] = m1[8] * m2[1] + m1[9] * m2[5] + m1[10] * m2[9] + \
                    m1[11] * m2[13]
        self.m[10] = m1[8] * m2[2] + m1[9] * m2[6] + m1[10] * m2[10] + \
                     m1[11] * m2[14]
        self.m[11] = m1[8] * m2[3] + m1[9] * m2[7] + m1[10] * m2[11] + \
                     m1[11] * m2[15]

        self.m[12] = m1[12] * m2[0] + m1[13] * m2[4] + m1[14] * m2[8] + \
                     m1[15] * m2[12]
        self.m[13] = m1[12] * m2[1] + m1[13] * m2[5] + m1[14] * m2[9] + \
                     m1[15] * m2[13]
        self.m[14] = m1[12] * m2[2] + m1[13] * m2[6] + m1[14] * m2[10] + \
                     m1[15] * m2[14]
        self.m[15] = m1[12] * m2[3] + m1[13] * m2[7] + m1[14] * m2[11] + \
                     m1[15] * m2[15]
        return self

    def __mul__(self, Matrix44 other) -> Matrix44:
        cdef Matrix44 res_matrix = self.copy()
        return res_matrix.__imul__(other)

    # __matmul__ = __mul__ does not work!

    def __matmul__(self, Matrix44 other) -> Matrix44:
        cdef Matrix44 res_matrix = self.copy()
        return res_matrix.__imul__(other)

    def transpose(self) -> None:
        swap(&self.m[1], &self.m[4])
        swap(&self.m[2], &self.m[8])
        swap(&self.m[3], &self.m[12])
        swap(&self.m[6], &self.m[9])
        swap(&self.m[7], &self.m[13])
        swap(&self.m[11], &self.m[14])

    def determinant(self) -> float:
        cdef double *m = self.m
        return m[0] * m[5] * m[10] * m[15] - m[0] * m[5] * m[11] * m[14] + \
               m[0] * m[6] * m[11] * m[13] - m[0] * m[6] * m[9] * m[15] + \
               m[0] * m[7] * m[9] * m[14] - m[0] * m[7] * m[10] * m[13] - \
               m[1] * m[6] * m[11] * m[12] + m[1] * m[6] * m[8] * m[15] - \
               m[1] * m[7] * m[8] * m[14] + m[1] * m[7] * m[10] * m[12] - \
               m[1] * m[4] * m[10] * m[15] + m[1] * m[4] * m[11] * m[14] + \
               m[2] * m[7] * m[8] * m[13] - m[2] * m[7] * m[9] * m[12] + \
               m[2] * m[4] * m[9] * m[15] - m[2] * m[4] * m[11] * m[13] + \
               m[2] * m[5] * m[11] * m[12] - m[2] * m[5] * m[8] * m[15] - \
               m[3] * m[4] * m[9] * m[14] + m[3] * m[4] * m[10] * m[13] - \
               m[3] * m[5] * m[10] * m[12] + m[3] * m[5] * m[8] * m[14] - \
               m[3] * m[6] * m[8] * m[13] + m[3] * m[6] * m[9] * m[12]

    def inverse(self) -> None:
        cdef double[16] m = self.m  # memcopy
        cdef double det = self.determinant()
        cdef double f = 1. / det
        self.m[0] = (m[6] * m[11] * m[13] - m[7] * m[10] * m[13] + m[7] * m[9] *
                     m[14] - m[5] * m[11] * m[14] - m[6] * m[9] * m[15] + m[5] *
                     m[10] * m[15]) * f

        self.m[1] = (m[3] * m[10] * m[13] - m[2] * m[11] * m[13] - m[3] * m[9] *
                     m[14] + m[1] * m[11] * m[14] + m[2] * m[9] * m[15] -
                     m[1] * m[10] * m[15]) * f
        self.m[2] = (m[2] * m[7] * m[13] - m[3] * m[6] * m[13] + m[3] * m[5] *
                     m[14] - m[1] * m[7] * m[14] - m[2] * m[5] * m[15] +
                     m[1] * m[6] * m[15]) * f

        self.m[3] = (m[3] * m[6] * m[9] - m[2] * m[7] * m[9] - m[3] * m[5] *
                     m[10] + m[1] * m[7] * m[10] + m[2] * m[5] * m[11] - m[1] *
                     m[6] * m[11]) * f

        self.m[4] = (m[7] * m[10] * m[12] - m[6] * m[11] * m[12] - m[7] * m[8] *
                     m[14] + m[4] * m[11] * m[14] + m[6] * m[8] * m[15] -
                     m[4] * m[10] * m[15]) * f

        self.m[5] = (m[2] * m[11] * m[12] - m[3] * m[10] * m[12] + m[3] * m[8] *
                     m[14] - m[0] * m[11] * m[14] - m[2] * m[8] * m[15] +
                     m[0] * m[10] * m[15]) * f

        self.m[6] = (m[3] * m[6] * m[12] - m[2] * m[7] * m[12] - m[3] * m[4] *
                     m[14] + m[0] * m[7] * m[14] + m[2] * m[4] * m[15] -
                     m[0] * m[6] * m[15]) * f

        self.m[7] = (m[2] * m[7] * m[8] - m[3] * m[6] * m[8] + m[3] * m[4] *
                     m[10] - m[0] * m[7] * m[10] - m[2] * m[4] * m[11] +
                     m[0] * m[6] * m[11]) * f

        self.m[8] = (m[5] * m[11] * m[12] - m[7] * m[9] * m[12] + m[7] * m[8] *
                     m[13] - m[4] * m[11] * m[13] - m[5] * m[8] * m[15] +
                     m[4] * m[9] * m[15]) * f

        self.m[9] = (m[3] * m[9] * m[12] - m[1] * m[11] * m[12] - m[3] *
                     m[8] * m[13] + m[0] * m[11] * m[13] + m[1] * m[8] *
                     m[15] - m[0] * m[9] * m[15]) * f

        self.m[10] = (m[1] * m[7] * m[12] - m[3] * m[5] * m[12] + m[3] *
                      m[4] * m[13] - m[0] * m[7] * m[13] - m[1] * m[4] *
                      m[15] + m[0] * m[5] * m[15]) * f

        self.m[11] = (m[3] * m[5] * m[8] - m[1] * m[7] * m[8] - m[3] * m[4] *
                      m[9] + m[0] * m[7] * m[9] + m[1] * m[4] * m[11] -
                      m[0] * m[5] * m[11]) * f

        self.m[12] = (m[6] * m[9] * m[12] - m[5] * m[10] * m[12] - m[6] *
                      m[8] * m[13] + m[4] * m[10] * m[13] + m[5] * m[8] *
                      m[14] - m[4] * m[9] * m[14]) * f

        self.m[13] = (m[1] * m[10] * m[12] - m[2] * m[9] * m[12] + m[2] *
                      m[8] * m[13] - m[0] * m[10] * m[13] - m[1] * m[8] *
                      m[14] + m[0] * m[9] * m[14]) * f

        self.m[14] = (m[2] * m[5] * m[12] - m[1] * m[6] * m[12] - m[2] *
                      m[4] * m[13] + m[0] * m[6] * m[13] + m[1] * m[4] *
                      m[14] - m[0] * m[5] * m[14]) * f

        self.m[15] = (m[1] * m[6] * m[8] - m[2] * m[5] * m[8] + m[2] * m[4] *
                      m[9] - m[0] * m[6] * m[9] - m[1] * m[4] * m[10] +
                      m[0] * m[5] * m[10]) * f

    @staticmethod
    def ucs(ux=X_AXIS, uy=Y_AXIS, uz=Z_AXIS, origin=NULLVEC) -> Matrix44:
        cdef Matrix44 mat = Matrix44()
        cdef Vec3 _ux = Vec3(ux)
        cdef Vec3 _uy = Vec3(uy)
        cdef Vec3 _uz = Vec3(uz)
        cdef Vec3 _origin = Vec3(origin)

        mat.m[0] = _ux.x
        mat.m[1] = _ux.y
        mat.m[2] = _ux.z

        mat.m[4] = _uy.x
        mat.m[5] = _uy.y
        mat.m[6] = _uy.z

        mat.m[8] = _uz.x
        mat.m[9] = _uz.y
        mat.m[10] = _uz.z

        mat.m[12] = _origin.x
        mat.m[13] = _origin.y
        mat.m[14] = _origin.z

        return mat

    def transform(self, vector: UVec) -> Vec3:
        cdef Vec3 res = Vec3(vector)
        cdef double x = res.x
        cdef double y = res.y
        cdef double z = res.z
        cdef double *m = self.m

        res.x = x * m[0] + y * m[4] + z * m[8] + m[12]
        res.y = x * m[1] + y * m[5] + z * m[9] + m[13]
        res.z = x * m[2] + y * m[6] + z * m[10] + m[14]
        return res

    def transform_direction(self, vector: UVec, normalize=False) -> Vec3:
        cdef Vec3 res = Vec3(vector)
        cdef double x = res.x
        cdef double y = res.y
        cdef double z = res.z
        cdef double *m = self.m

        res.x = x * m[0] + y * m[4] + z * m[8]
        res.y = x * m[1] + y * m[5] + z * m[9]
        res.z = x * m[2] + y * m[6] + z * m[10]
        if normalize:
            return v3_normalize(res, 1.0)
        else:
            return res

    ocs_to_wcs = transform_direction

    def transform_vertices(self, vectors: Iterable[UVec]) -> Iterator[Vec3]:
        cdef double *m = self.m
        cdef Vec3 res
        cdef double x, y, z

        for vector in vectors:
            res = Vec3(vector)
            x = res.x
            y = res.y
            z = res.z

            res.x = x * m[0] + y * m[4] + z * m[8] + m[12]
            res.y = x * m[1] + y * m[5] + z * m[9] + m[13]
            res.z = x * m[2] + y * m[6] + z * m[10] + m[14]
            yield res

    def fast_2d_transform(self, points: Iterable[UVec]) -> Iterator[Vec2]:
        cdef double m0 = self.m[0]
        cdef double m1 = self.m[1]
        cdef double m4 = self.m[4]
        cdef double m5 = self.m[5]
        cdef double m12 = self.m[12]
        cdef double m13 = self.m[13]
        cdef double x, y
        cdef Vec2 res

        for pnt in points:
            res = Vec2(pnt)
            x = res.x
            y = res.y
            res.x = x * m0 + y * m4 + m12
            res.y = x * m1 + y * m5 + m13
            yield res

    def transform_array_inplace(self, array: np.ndarray, ndim: int) -> None:
        """Transforms a numpy array inplace, the argument `ndim` defines the dimensions
        to transform, this allows 2D/3D transformation on arrays with more columns
        e.g. a polyline array which stores points as (x, y, start_width, end_width,
        bulge) values.

        """
        cdef int _ndim = ndim
        if _ndim == 2:
            assert array.shape[1] > 1
            transform_2d_array_inplace(self.m, array, array.shape[0])
        elif _ndim == 3:
            assert array.shape[1] > 2
            transform_3d_array_inplace(self.m, array, array.shape[0])
        else:
            raise ValueError("ndim has to be 2 or 3")


    def transform_directions(
        self, vectors: Iterable[UVec], normalize=False
    ) -> Iterator[Vec3]:
        cdef double *m = self.m
        cdef Vec3 res
        cdef double x, y, z
        cdef bint _normalize = normalize

        for vector in vectors:
            res = Vec3(vector)
            x = res.x
            y = res.y
            z = res.z

            res.x = x * m[0] + y * m[4] + z * m[8]
            res.y = x * m[1] + y * m[5] + z * m[9]
            res.z = x * m[2] + y * m[6] + z * m[10]
            yield v3_normalize(res, 1.0) if _normalize else res

    def ucs_vertex_from_wcs(self, wcs: Vec3) -> Vec3:
        return self.ucs_direction_from_wcs(wcs - self.origin)

    def ucs_direction_from_wcs(self, wcs: UVec) -> Vec3:
        cdef double *m = self.m
        cdef Vec3 res = Vec3(wcs)
        cdef double x = res.x
        cdef double y = res.y
        cdef double z = res.z

        res.x = x * m[0] + y * m[1] + z * m[2]
        res.y = x * m[4] + y * m[5] + z * m[6]
        res.z = x * m[8] + y * m[9] + z * m[10]
        return res

    ocs_from_wcs = ucs_direction_from_wcs


@cython.boundscheck(False)
@cython.wraparound(False)
cdef void transform_2d_array_inplace(double *m, double [:, ::1] array, Py_ssize_t size):
    cdef double m0 = m[0]
    cdef double m1 = m[1]
    cdef double m4 = m[4]
    cdef double m5 = m[5]
    cdef double m12 = m[12]
    cdef double m13 = m[13]
    cdef double x, y
    cdef Py_ssize_t i

    for i in range(size):
        x = array[i, 0]
        y = array[i, 1]
        array[i, 0] = x * m0 + y * m4 + m12
        array[i, 1] = x * m1 + y * m5 + m13

@cython.boundscheck(False)
@cython.wraparound(False)
cdef void transform_3d_array_inplace(double *m, double [:, ::1] array, Py_ssize_t size):
    cdef double x, y, z
    cdef Py_ssize_t i

    for i in range(size):
        x = array[i, 0]
        y = array[i, 1]
        z = array[i, 2]

        array[i, 0] = x * m[0] + y * m[4] + z * m[8] + m[12]
        array[i, 1] = x * m[1] + y * m[5] + z * m[9] + m[13]
        array[i, 2] = x * m[2] + y * m[6] + z * m[10] + m[14]



# Copyright (c) 2020-2024, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import TYPE_CHECKING, Iterable, Callable, Optional
from typing_extensions import Self

from ezdxf.entities import factory, DXFGraphic, SeqEnd, DXFEntity
from ezdxf.lldxf import const
from .copy import default_copy
import logging

if TYPE_CHECKING:
    from ezdxf.document import Drawing
    from ezdxf.entities import DXFEntity
    from ezdxf.entitydb import EntityDB
    from ezdxf import xref


__all__ = ["entity_linker", "LinkedEntities"]
logger = logging.getLogger("ezdxf")


class LinkedEntities(DXFGraphic):
    """Super class for common features of the INSERT and the POLYLINE entity.
    Both have linked entities like the VERTEX or ATTRIB entity and a
    SEQEND entity.

    """

    def __init__(self) -> None:
        super().__init__()
        self._sub_entities: list[DXFGraphic] = []
        self.seqend: Optional[SeqEnd] = None

    def copy_data(self, entity: Self, copy_strategy=default_copy) -> None:
        """Copy all sub-entities ands SEQEND. (internal API)"""
        assert isinstance(entity, LinkedEntities)
        entity._sub_entities = [
            copy_strategy.copy(e) for e in self._sub_entities
        ]
        if self.seqend:
            entity.seqend = copy_strategy.copy(self.seqend)

    def link_entity(self, entity: DXFEntity) -> None:
        """Link VERTEX to ATTRIB entities."""
        assert isinstance(entity, DXFGraphic)
        entity.set_owner(self.dxf.handle, self.dxf.paperspace)
        self._sub_entities.append(entity)

    def link_seqend(self, seqend: DXFEntity) -> None:
        """Link SEQEND entity. (internal API)"""
        seqend.dxf.owner = self.dxf.owner
        if self.seqend is not None:  # destroy existing seqend
            self.seqend.destroy()
        self.seqend = seqend  # type: ignore

    def post_bind_hook(self):
        """Create always a SEQEND entity."""
        if self.seqend is None:
            self.new_seqend()

    def all_sub_entities(self) -> Iterable[DXFEntity]:
        """Yields all sub-entities and SEQEND. (internal API)"""
        yield from self._sub_entities
        if self.seqend:
            yield self.seqend

    def process_sub_entities(self, func: Callable[[DXFEntity], None]):
        """Call `func` for all sub-entities and SEQEND. (internal API)"""
        for entity in self.all_sub_entities():
            if entity.is_alive:
                func(entity)

    def add_sub_entities_to_entitydb(self, db: EntityDB) -> None:
        """Add sub-entities (VERTEX, ATTRIB, SEQEND) to entity database `db`,
        called from EntityDB. (internal API)
        """

        def add(entity: DXFEntity):
            entity.doc = self.doc  # grant same document
            db.add(entity)

        if not self.seqend or not self.seqend.is_alive:
            self.new_seqend()
        self.process_sub_entities(add)

    def new_seqend(self):
        """Create and bind new SEQEND. (internal API)"""
        attribs = {"layer": self.dxf.layer}
        if self.doc:
            seqend = factory.create_db_entry("SEQEND", attribs, self.doc)
        else:
            seqend = factory.new("SEQEND", attribs)
        self.link_seqend(seqend)

    def set_owner(self, owner: Optional[str], paperspace: int = 0):
        """Set owner of all sub-entities and SEQEND. (internal API)"""
        # Loading from file: POLYLINE/INSERT will be added to layout before
        # vertices/attrib entities are linked, so set_owner() of POLYLINE does
        # not set owner of vertices at loading time.
        super().set_owner(owner, paperspace)
        self.take_ownership()

    def take_ownership(self):
        """Take ownership of all sub-entities and SEQEND. (internal API)"""
        handle = self.dxf.handle
        paperspace = self.dxf.paperspace
        for entity in self.all_sub_entities():
            if entity.is_alive:
                entity.dxf.owner = handle
                entity.dxf.paperspace = paperspace

    def remove_dependencies(self, other: Optional[Drawing] = None):
        """Remove all dependencies from current document to bind entity to
        `other` document. (internal API)
        """
        self.process_sub_entities(lambda e: e.remove_dependencies(other))
        super().remove_dependencies(other)

    def destroy(self) -> None:
        """Destroy all data and references."""
        if not self.is_alive:
            return

        self.process_sub_entities(func=lambda e: e.destroy())
        del self._sub_entities
        del self.seqend
        super().destroy()

    def register_resources(self, registry: xref.Registry) -> None:
        """Register required resources to the resource registry."""
        super().register_resources(registry)
        self.process_sub_entities(lambda e: e.register_resources(registry))

    def map_resources(self, clone: Self, mapping: xref.ResourceMapper) -> None:
        """Translate resources from self to the copied entity."""
        assert isinstance(clone, LinkedEntities)
        super().map_resources(clone, mapping)
        for source, _clone in zip(self.all_sub_entities(), clone.all_sub_entities()):
            source.map_resources(_clone, mapping)


LINKED_ENTITIES = {"INSERT": "ATTRIB", "POLYLINE": "VERTEX"}


def entity_linker() -> Callable[[DXFEntity], bool]:
    """Create an DXF entities linker."""
    main_entity: Optional[DXFEntity] = None
    expected_dxftype = ""

    def entity_linker_(entity: DXFEntity) -> bool:
        """Collect and link entities which are linked to a parent entity:

        - VERTEX -> POLYLINE
        - ATTRIB -> INSERT

        Args:
             entity: examined DXF entity

        Returns:
             True if `entity` is linked to a parent entity

        """
        nonlocal main_entity, expected_dxftype
        dxftype: str = entity.dxftype()
        # INSERT & POLYLINE are not linked entities, they are stored in the
        # entity space.
        are_linked_entities = False
        if main_entity is not None:
            # VERTEX, ATTRIB & SEQEND are linked tags, they are NOT stored in
            # the entity space.
            are_linked_entities = True
            if dxftype == "SEQEND":
                main_entity.link_seqend(entity)  # type: ignore
                # Marks also the end of the main entity
                main_entity = None
            # Check for valid DXF structure:
            #   VERTEX follows POLYLINE
            #   ATTRIB follows INSERT
            elif dxftype == expected_dxftype:
                main_entity.link_entity(entity)  # type: ignore
            else:
                raise const.DXFStructureError(
                    f"Expected DXF entity {dxftype} or SEQEND"
                )

        elif dxftype in LINKED_ENTITIES:
            # Only INSERT and POLYLINE have a linked entities structure:
            if dxftype == "INSERT" and not entity.dxf.get("attribs_follow", 0):
                # INSERT must not have following ATTRIBS:
                #
                #   INSERT with no ATTRIBS, attribs_follow == 0
                #   ATTRIB as a stand alone entity, which is a DXF structure
                #   error, but this error should be handled in the audit
                #   process.
                #   ....
                #   INSERT with ATTRIBS, attribs_follow == 1
                #   ATTRIB as connected entity
                #   SEQEND
                #
                # Therefore a ATTRIB following an INSERT doesn't mean that
                # these entities are linked.
                pass
            else:
                main_entity = entity
                expected_dxftype = LINKED_ENTITIES[dxftype]

        # Attached MTEXT entity - this feature most likely does not exist!
        elif (dxftype == "MTEXT") and (entity.dxf.handle is None):
            logger.error(
                "Found attached MTEXT entity. Please open an issue at github: "
                "https://github.com/mozman/ezdxf/issues and provide a DXF "
                "example file."
            )
        return are_linked_entities

    return entity_linker_



# Copyright (c) 2011-2024, Manfred Moitzi
# License: MIT License
import pytest
import ezdxf
from ezdxf.lldxf.const import VTX_3D_POLYLINE_VERTEX
from ezdxf.lldxf.tagwriter import TagCollector
from ezdxf.tools.test import load_entities
from ezdxf.sections.entities import EntitySection
from ezdxf.entities import factory, Polyline


@pytest.fixture(scope="module")
def doc():
    return ezdxf.new("R2000")


@pytest.fixture(scope="module")
def layout(doc):
    return doc.modelspace()


def test_create_polyline2D(layout):
    polyline = layout.add_polyline2d([(0, 0), (1, 1)])
    assert (0.0, 0.0) == polyline[0].dxf.location
    assert (1.0, 1.0) == polyline[1].dxf.location
    assert "AcDb2dPolyline" == polyline.get_mode()


def test_create_polyline3D(layout):
    polyline = layout.add_polyline3d([(1, 2, 3), (4, 5, 6)])
    assert (1.0, 2.0, 3.0) == polyline[0].dxf.location
    assert (4.0, 5.0, 6.0) == polyline[1].dxf.location
    assert VTX_3D_POLYLINE_VERTEX == polyline[0].dxf.flags
    assert "AcDb3dPolyline" == polyline.get_mode()


def test_add_new_sub_entities_to_entity_database(layout, doc):
    db = doc.entitydb
    db_len = len(db)

    polyline = layout.add_polyline2d([(0, 0), (1, 1)])
    assert len(db) == db_len + 4  # POLYLINE-VERTEX-VERTEX-SEQEND
    assert polyline.seqend.is_alive is True

    db_len = len(db)
    polyline.append_vertices([(2, 2), (3, 3)])
    assert len(polyline) == 4
    assert polyline.vertices[-1].dxf.handle is not None
    assert len(db) == db_len + 2, "new VERTEX entities should be added to db"


def test_export_sub_entities_to_dxf(layout, doc):
    from ezdxf.lldxf.tagwriter import TagCollector

    layout.add_polyline2d([(0, 0), (1, 1)])
    writer = TagCollector()
    layout.entity_space.export_dxf(tagwriter=writer)
    structure_tags = [tag for tag in writer.tags if tag[0] == 0]
    assert structure_tags[0] == (0, "POLYLINE")
    assert structure_tags[1] == (0, "VERTEX")
    assert structure_tags[2] == (0, "VERTEX")
    assert structure_tags[3] == (0, "SEQEND")


def test_vertex_layer(layout):
    attribs = {"layer": "polyline_layer"}
    polyline = layout.add_polyline3d([(1, 2, 3), (4, 5, 6)], dxfattribs=attribs)
    for vertex in polyline.vertices:
        assert (
            "polyline_layer" == vertex.dxf.layer
        ), "VERTEX entity not on the same layer as the POLYLINE entity."


def test_change_polyline_layer(layout):
    attribs = {"layer": "polyline_layer"}
    polyline = layout.add_polyline3d([(1, 2, 3), (4, 5, 6)], dxfattribs=attribs)
    polyline.dxf.layer = "changed_layer"
    for vertex in polyline.vertices:
        assert (
            "changed_layer" == vertex.dxf.layer
        ), "VERTEX entity not on the same layer as the POLYLINE entity."


def test_set_vertex(layout):
    polyline = layout.add_polyline2d([(0, 0), (1, 1), (2, 2), (3, 3)])
    polyline[2].dxf.location = (7, 7)
    assert (7.0, 7.0) == polyline[2].dxf.location


def test_points(layout):
    points = [(0, 0), (1, 1), (2, 2), (3, 3)]
    polyline = layout.add_polyline2d(points)
    assert points == list(polyline.points())


def test_points_in_wcs(layout):
    points = [(0, 0), (1, 1), (2, 2), (3, 3)]
    polyline = layout.add_polyline2d(points)
    assert points == list(polyline.points_in_wcs())


def test_points_in_wcs_with_elevation(layout):
    points = [(0, 0), (1, 1), (2, 2), (3, 3)]
    polyline = layout.add_polyline2d(points)
    polyline.dxf.elevation = (0, 0, 2)
    expected = [(p[0], p[1], 2) for p in points]
    assert expected == list(polyline.points_in_wcs())


def test_points_in_wcs_with_ocs(layout):
    points = [(0, 0), (1, 1), (2, 2), (3, 3)]
    polyline = layout.add_polyline2d(points)
    polyline.dxf.elevation = (0, 0, 7)
    polyline.dxf.extrusion = (0, 0, -1)
    expected = [(0, 0, -7), (-1, 1, -7), (-2, 2, -7), (-3, 3, -7)]
    assert expected == list(polyline.points_in_wcs())


def test_point_slicing(layout):
    points = [(0, 0), (1, 1), (2, 2), (3, 3)]
    polyline = layout.add_polyline2d(points)
    assert [(1, 1), (2, 2)] == list(polyline.points())[1:3]


def test_append_vertices(layout):
    polyline = layout.add_polyline2d([(0, 0), (1, 1)])
    polyline.append_vertices([(7, 7), (8, 8)])
    assert (7.0, 7.0) == polyline[2].dxf.location
    assert 4 == len(polyline)


def test_insert_vertices(layout):
    polyline = layout.add_polyline2d([(0, 0), (1, 1)])
    polyline.insert_vertices(0, [(7, 7), (8, 8)])
    assert (7, 7) == polyline[0].dxf.location
    assert (1, 1) == polyline[3].dxf.location
    assert 4 == len(polyline)


def test_delete_one_vertex(layout):
    polyline = layout.add_polyline2d([(0, 0), (1, 1), (2, 2), (3, 3)])
    del polyline.vertices[0]
    assert (1, 1) == polyline[0].dxf.location
    assert 3 == len(polyline)


def test_delete_two_vertices(layout):
    polyline = layout.add_polyline2d([(0, 0), (1, 1), (2, 2), (3, 3)])
    del polyline.vertices[0:2]
    assert (2, 2) == polyline[0].dxf.location
    assert 2 == len(polyline)


def test_polymesh_correct_casting(layout):
    mesh1 = layout.add_polymesh((4, 4))
    assert mesh1.dxf.handle is not None
    mesh2 = layout[-1]
    assert mesh1 is mesh2
    assert mesh1.vertices is mesh2.vertices


def test_polymesh_set_vertex(layout):
    mesh = layout.add_polymesh((4, 4))
    mesh.set_mesh_vertex((1, 1), (1, 2, 3))
    assert (1, 2, 3) == mesh.get_mesh_vertex((1, 1)).dxf.location


def test_polymesh_error_n_index(layout):
    mesh = layout.add_polymesh((4, 4))
    with pytest.raises(IndexError):
        mesh.get_mesh_vertex((0, 4))


def test_polymesh_error_m_index(layout):
    mesh = layout.add_polymesh((4, 4))
    with pytest.raises(IndexError):
        mesh.get_mesh_vertex((4, 0))


def test_polymesh_mesh_cache(layout):
    pos = (2, 1)
    mesh = layout.add_polymesh((4, 4))
    cache = mesh.get_mesh_vertex_cache()
    cache[pos] = (1, 2, 3)
    vertex = mesh.get_mesh_vertex(pos)
    assert vertex.dxf.location == cache[pos]
    with pytest.raises(ezdxf.DXFIndexError):
        cache[4, 0]


def test_polyface_correct_casting(layout):
    polyface1 = layout.add_polyface()
    polyface2 = layout[-1]
    assert polyface1 is polyface2
    assert polyface1.vertices is polyface2.vertices


def test_polyface_create_face(layout):
    face = layout.add_polyface()
    assert 0 == len(face)


def test_polyface_add_face(layout):
    face = layout.add_polyface()
    face.append_face([(0, 0), (1, 1), (2, 2), (3, 3)])
    assert [(0, 0), (1, 1) == (2, 2), (3, 3), (0, 0, 0)], list(face.points())


def test_polyface_face_indices(layout):
    face = layout.add_polyface()
    face.append_face([(0, 0), (1, 1), (2, 2), (3, 3)])
    face_record = face[4]
    assert 1 == face_record.dxf.vtx0
    assert 2 == face_record.dxf.vtx1
    assert 3 == face_record.dxf.vtx2
    assert 4 == face_record.dxf.vtx3


def test_polyface_add_two_face_indices(layout):
    face = layout.add_polyface()
    face.append_face([(0, 0), (1, 1), (2, 2), (3, 3)])
    # second face has same vertices as the first face
    face.append_face([(0, 0), (1, 1), (2, 2)])
    face_record = face[5]  # second face
    assert 1 == face_record.dxf.vtx0
    assert 2 == face_record.dxf.vtx1
    assert 3 == face_record.dxf.vtx2
    assert 4 == face.dxf.m_count  # vertices count
    assert 2 == face.dxf.n_count  # faces count


def test_polyface_faces(layout):
    face = layout.add_polyface()
    face.append_face([(0, 0), (1, 1), (2, 2), (3, 3)])
    face.append_face([(0, 0), (1, 1), (2, 2)])
    result = list(face.faces())
    assert 2 == len(result)
    points1 = [vertex.dxf.location for vertex in result[0]]
    # the last vertex is the face_record and is always (0,0,0)
    # the face_record contains indices to the face building vertices
    assert [(0, 0), (1, 1), (2, 2), (3, 3), (0, 0, 0)] == points1


def test_polyface_optimized_cube(layout):
    face = layout.add_polyface()
    # a cube consist of 6 faces and 24 vertices
    # duplicated vertices should be removed
    face.append_faces(cube_faces())
    assert 8 == face.dxf.m_count  # vertices count
    assert 6 == face.dxf.n_count  # faces count


def cube_faces():
    # cube corner points
    p1 = (0, 0, 0)
    p2 = (0, 0, 1)
    p3 = (0, 1, 0)
    p4 = (0, 1, 1)
    p5 = (1, 0, 0)
    p6 = (1, 0, 1)
    p7 = (1, 1, 0)
    p8 = (1, 1, 1)

    # define the 6 cube faces
    # look into -x direction
    # Every add_face adds 4 vertices 6x4 = 24 vertices
    return [
        [p1, p5, p7, p3],
        [p1, p5, p6, p2],
        [p5, p7, p8, p6],
        [p7, p8, p4, p3],
        [p1, p3, p4, p2],
        [p2, p6, p8, p4],
    ]


def test_export_polyline2d(layout):
    polyline = layout.add_polyline2d([(0, 0), (1, 1)])
    collector = TagCollector()
    polyline.export_dxf(collector)
    tags = collector.tags
    assert (100, "AcDb2dPolyline") in tags
    collector.reset()
    polyline[0].export_dxf(collector)
    tags = collector.tags
    assert (100, "AcDbVertex") in tags
    assert (100, "AcDb2dVertex") in tags


def test_export_polyline3d(layout):
    polyline = layout.add_polyline3d([(0, 0), (1, 1)])
    collector = TagCollector()
    polyline.export_dxf(collector)
    tags = collector.tags
    assert (100, "AcDb3dPolyline") in tags
    collector.reset()
    polyline[0].export_dxf(collector)
    tags = collector.tags
    assert (100, "AcDbVertex") in tags
    assert (100, "AcDb3dPolylineVertex") in tags


def test_internals_polymesh(layout):
    mesh = layout.add_polymesh((4, 4))
    collector = TagCollector()
    mesh.export_dxf(collector)
    tags = collector.tags
    assert (100, "AcDbPolygonMesh") in tags
    collector.reset()
    mesh[0].export_dxf(collector)
    tags = collector.tags
    assert (100, "AcDbVertex") in tags
    assert (100, "AcDbPolygonMeshVertex") in tags


def test_internals_polyface(layout):
    face = layout.add_polyface()
    face.append_face([(0, 0), (1, 1), (2, 2), (3, 3)])

    collector = TagCollector()
    face.export_dxf(collector)
    tags = collector.tags
    assert (100, "AcDbPolyFaceMesh") in tags

    collector.reset()
    face[0].export_dxf(collector)
    tags = collector.tags
    assert (100, "AcDbVertex") in tags
    assert (100, "AcDbPolyFaceMeshVertex") in tags

    collector.reset()
    face[4].export_dxf(collector)
    tags = collector.tags
    assert (100, "AcDbVertex") not in tags
    assert (100, "AcDbFaceRecord") in tags


def test_new_style_polyface_face_count():
    doc = ezdxf.new()
    section = EntitySection(doc, load_entities(NEW_STYLE_POLYFACE, "ENTITIES"))
    entities = list(section)
    polyface = entities[0]
    faces = list(polyface.faces())
    assert 6 == len(faces)


def test_add_virtual_polyline_to_layout(doc, layout):
    polyline = Polyline()
    polyline.append_vertex((0, 0))
    layout.add_entity(polyline)
    assert factory.is_bound(polyline, doc) is True, "POLYLINE must be bound to document"
    assert (
        factory.is_bound(polyline.seqend, doc) is True
    ), "SEQEND must be bound to document"

    assert len(polyline) == 1
    assert (
        factory.is_bound(polyline.vertices[0], doc) is True
    ), "VERTEX must be bound to document"


NEW_STYLE_POLYFACE = """  0
SECTION
  2
ENTITIES
  0
POLYLINE
  5
9A
330
6B
100
AcDbEntity
  8
0
100
AcDbPolyFaceMesh
 66
     1
 10
0.0
 20
0.0
 30
0.0
 70
    64
 71
     8
 72
     6
  0
VERTEX
  5
9B
330
9A
100
AcDbEntity
  8
0
100
AcDbVertex
100
AcDbPolyFaceMeshVertex
 10
0.0
 20
0.0
 30
0.802929163112954
 70
   192
  0
VERTEX
  5
9C
330
9A
100
AcDbEntity
  8
0
100
AcDbVertex
100
AcDbPolyFaceMeshVertex
 10
0.4434395109072581
 20
0.0
 30
0.802929163112954
 70
   192
  0
VERTEX
  5
9D
330
9A
100
AcDbEntity
  8
0
100
AcDbVertex
100
AcDbPolyFaceMeshVertex
 10
0.4434395109072581
 20
0.4434395109072581
 30
0.802929163112954
 70
   192
  0
VERTEX
  5
9E
330
9A
100
AcDbEntity
  8
0
100
AcDbVertex
100
AcDbPolyFaceMeshVertex
 10
0.0
 20
0.4434395109072581
 30
0.802929163112954
 70
   192
  0
VERTEX
  5
9F
330
9A
100
AcDbEntity
  8
0
100
AcDbVertex
100
AcDbPolyFaceMeshVertex
 10
0.4434395109072581
 20
0.0
 30
1.246368674020211
 70
   192
  0
VERTEX
  5
A0
330
9A
100
AcDbEntity
  8
0
100
AcDbVertex
100
AcDbPolyFaceMeshVertex
 10
0.0
 20
0.0
 30
1.246368674020211
 70
   192
  0
VERTEX
  5
A1
330
9A
100
AcDbEntity
  8
0
100
AcDbVertex
100
AcDbPolyFaceMeshVertex
 10
0.4434395109072581
 20
0.4434395109072581
 30
1.246368674020211
 70
   192
  0
VERTEX
  5
A2
330
9A
100
AcDbEntity
  8
0
100
AcDbVertex
100
AcDbPolyFaceMeshVertex
 10
0.0
 20
0.4434395109072581
 30
1.246368674020211
 70
   192
  0
VERTEX
  5
A3
330
9A
100
AcDbEntity
  8
0
 62
     1
100
AcDbFaceRecord
 10
0.0
 20
0.0
 30
0.0
 70
   128
 71
     1
 72
     2
 73
     3
 74
     4
  0
VERTEX
  5
A4
330
9A
100
AcDbEntity
  8
0
 62
     2
100
AcDbFaceRecord
 10
0.0
 20
0.0
 30
0.0
 70
   128
 71
     1
 72
     2
 73
     5
 74
     6
  0
VERTEX
  5
A5
330
9A
100
AcDbEntity
  8
0
 62
     3
100
AcDbFaceRecord
 10
0.0
 20
0.0
 30
0.0
 70
   128
 71
     2
 72
     3
 73
     7
 74
     5
  0
VERTEX
  5
A6
330
9A
100
AcDbEntity
  8
0
 62
     4
100
AcDbFaceRecord
 10
0.0
 20
0.0
 30
0.0
 70
   128
 71
     3
 72
     7
 73
     8
 74
     4
  0
VERTEX
  5
A7
330
9A
100
AcDbEntity
  8
0
 62
     5
100
AcDbFaceRecord
 10
0.0
 20
0.0
 30
0.0
 70
   128
 71
     1
 72
     4
 73
     8
 74
     6
  0
VERTEX
  5
A8
330
9A
100
AcDbEntity
  8
0
 62
     6
100
AcDbFaceRecord
 10
0.0
 20
0.0
 30
0.0
 70
   128
 71
     6
 72
     5
 73
     7
 74
     8
  0
SEQEND
  5
A9
330
9A
100
AcDbEntity
  8
0
  0
ENDSEC
"""



# Copyright (c) 2019-2024 Manfred Moitzi
# License: MIT License
import pytest
import ezdxf
from ezdxf.audit import Auditor, AuditError, BlockCycleDetector
from ezdxf.entities import factory, DXFTagStorage, Attrib


@pytest.fixture(scope="module")
def doc():
    return ezdxf.new("R2000")


@pytest.fixture
def auditor(doc):
    return Auditor(doc)


@pytest.fixture
def entity(doc):
    return doc.modelspace().add_line((0, 0), (100, 0))


def test_color_index(entity, auditor):
    entity.dxf.__dict__["color"] = -1  # by pass 'set' validator
    auditor.check_entity_color_index(entity)
    assert len(auditor.fixes) == 1
    assert auditor.fixes[0].code == AuditError.INVALID_COLOR_INDEX

    auditor.reset()
    entity.dxf.__dict__["color"] = 258  # by pass 'set' validator
    auditor.check_entity_color_index(entity)
    assert len(auditor.fixes) == 1
    assert auditor.fixes[0].code == AuditError.INVALID_COLOR_INDEX


def test_lineweight_too_small(entity, auditor):
    entity.dxf.__dict__["lineweight"] = -5  # by pass 'set' validator
    auditor.check_entity_lineweight(entity)
    assert len(auditor.fixes) == 1
    assert auditor.fixes[0].code == AuditError.INVALID_LINEWEIGHT
    assert entity.dxf.lineweight == -1


def test_lineweight_too_big(entity, auditor):
    entity.dxf.__dict__["lineweight"] = 212  # by pass 'set' validator
    auditor.check_entity_lineweight(entity)
    assert len(auditor.fixes) == 1
    assert auditor.fixes[0].code == AuditError.INVALID_LINEWEIGHT
    assert entity.dxf.lineweight == 211


def test_invalid_lineweight(entity, auditor):
    entity.dxf.__dict__["lineweight"] = 10  # by pass 'set' validator
    auditor.check_entity_lineweight(entity)
    assert len(auditor.fixes) == 1
    assert auditor.fixes[0].code == AuditError.INVALID_LINEWEIGHT
    assert entity.dxf.lineweight == 13


def test_for_valid_layer_name(entity, auditor):
    entity.dxf.__dict__["layer"] = "Invalid/"  # by pass 'set' validator
    auditor.check_for_valid_layer_name(entity)
    assert len(auditor) == 1
    assert auditor.errors[0].code == AuditError.INVALID_LAYER_NAME


def test_if_layer_linetype_exist():
    doc = ezdxf.new()
    doc.layers.add("Layer0", linetype="DoesNotExist")
    auditor = doc.audit()
    assert len(auditor.fixes) == 1
    fix = auditor.fixes[0]
    assert fix.code == AuditError.UNDEFINED_LINETYPE


def test_for_existing_owner(entity, auditor):
    entity.dxf.owner = "FFFFFF"
    auditor.check_owner_exist(entity)
    auditor.empty_trashcan()
    assert len(auditor.fixes) == 1
    assert auditor.fixes[0].code == AuditError.INVALID_OWNER_HANDLE
    assert entity.is_alive is False, "delete entity without valid owner"


@pytest.mark.parametrize("TYPE", ("TEXT", "MTEXT", "ATTRIB", "ATTDEF"))
def test_for_existing_text_style(TYPE, auditor, doc):
    text = factory.new(TYPE, dxfattribs={"style": "UNDEFINED"}, doc=doc)
    auditor.check_text_style(text)
    assert len(auditor.fixes) == 1
    assert auditor.fixes[0].code == AuditError.UNDEFINED_TEXT_STYLE
    assert text.dxf.style == "Standard"


def test_block_cycle_detector_setup():
    doc = ezdxf.new()
    a = doc.blocks.new("a")
    b = doc.blocks.new("b")
    c = doc.blocks.new("c")
    a.add_blockref("b", (0, 0))
    a.add_blockref("c", (0, 0))
    b.add_blockref("c", (0, 0))
    c.add_blockref("a", (0, 0))  # cycle

    detector = BlockCycleDetector(doc)
    assert detector.has_cycle("a") is True
    assert detector.has_cycle("b") is True
    assert detector.has_cycle("c") is True

    auditor = Auditor(doc)
    auditor.check_block_reference_cycles()
    assert len(auditor.errors) == 3, "one entry for each involved block: 'a', 'b', 'c'"
    assert auditor.errors[0].code == AuditError.INVALID_BLOCK_REFERENCE_CYCLE
    assert auditor.errors[1].code == AuditError.INVALID_BLOCK_REFERENCE_CYCLE
    assert auditor.errors[2].code == AuditError.INVALID_BLOCK_REFERENCE_CYCLE


def test_block_cycle_detector(doc):
    detector = BlockCycleDetector(doc)
    data = {
        "a": set("bcd"),  # no cycle
        "b": set(),
        "c": set("x"),
        "d": set("xy"),
        "e": set("e"),  # short cycle
        "f": set("g"),
        "g": set("h"),
        "h": set("i"),
        "i": set("f"),  # long cycle
        "j": set("k"),
        "k": set("j"),  # short cycle
        "x": set(),
        "y": set(),
    }
    detector.blocks = data
    assert detector.has_cycle("a") is False
    assert detector.has_cycle("b") is False
    assert detector.has_cycle("c") is False
    assert detector.has_cycle("d") is False
    assert detector.has_cycle("e") is True
    assert detector.has_cycle("f") is True
    assert detector.has_cycle("g") is True
    assert detector.has_cycle("h") is True
    assert detector.has_cycle("i") is True
    assert detector.has_cycle("j") is True
    assert detector.has_cycle("k") is True
    assert detector.has_cycle("x") is False
    assert detector.has_cycle("y") is False


def test_broken_block_cycle_detector(doc):
    detector = BlockCycleDetector(doc)
    data = {
        "a": set("bcd"),  # 'd' does not exist
        "b": set(),
        "c": set(),
    }
    detector.blocks = data
    assert detector.has_cycle("a") is False
    assert detector.has_cycle("b") is False


def test_fix_invalid_leader(doc, auditor):
    msp = doc.modelspace()
    # no creator interface for LEADER (yet)
    leader = factory.new("LEADER", doc=doc)
    doc.entitydb.add(leader)
    msp.add_entity(leader)
    assert leader.is_alive is True

    leader.audit(auditor)
    assert leader.is_alive is False
    assert auditor.fixes[-1].code == AuditError.INVALID_VERTEX_COUNT


def test_fix_invalid_insert(doc, auditor):
    msp = doc.modelspace()
    insert = msp.add_blockref("TEST_INVALID_INSERT", (0, 0))
    insert.audit(auditor)
    auditor.empty_trashcan()  # explicit call required
    assert insert.is_alive is False
    assert auditor.fixes[-1].code == AuditError.UNDEFINED_BLOCK


def test_fix_insert_scale(doc, auditor):
    msp = doc.modelspace()
    test_block = "TEST_INSERT"
    if test_block not in doc.blocks:
        doc.blocks.new(test_block)
    insert = msp.add_blockref(
        test_block, (0, 0), dxfattribs={"xscale": 0, "yscale": 0, "zscale": 0}
    )
    insert.audit(auditor)
    assert insert.dxf.xscale == 1.0
    assert insert.dxf.xscale == 1.0
    assert insert.dxf.xscale == 1.0


def test_remove_invalid_entities_from_blocks():
    # The model space is just a BLOCK!
    doc = ezdxf.new()
    msp = doc.modelspace()
    # hack hack hack!
    msp.entity_space.add(DXFTagStorage())
    auditor = doc.audit()
    assert len(list(msp)) == 0
    assert len(auditor.fixes) == 1


def test_remove_standalone_attrib_entities_from_blocks():
    # The model space is just a BLOCK!
    doc = ezdxf.new()
    msp = doc.modelspace()
    # Missing tag is a different issue!
    msp.add_entity(Attrib.new(dxfattribs={"tag": "test"}))
    auditor = doc.audit()
    assert len(list(msp)) == 0
    assert len(auditor.fixes) == 1


def test_fix_invalid_transparency():
    doc = ezdxf.new()
    msp = doc.modelspace()
    line = msp.add_line((0, 0), (1, 0))
    # transparency value requires 0x02000000 bit set
    line.dxf.unprotected_set("transparency", 0x10000000)
    auditor = Auditor(doc)
    line.audit(auditor)
    assert line.dxf.hasattr("transparency") is False
    assert len(auditor.fixes) == 1


def test_fix_entities_with_invalid_owner_handle():
    doc = ezdxf.new()
    msp = doc.modelspace()
    line = msp.add_line((0, 0), (1, 0))

    block = doc.blocks.new("TEST")
    block.add_entity(line)

    result = doc.audit()
    fix = result.fixes[0]
    assert fix.code == AuditError.REMOVED_ENTITY_WITH_INVALID_OWNER_HANDLE

    assert len(block) == 1
    assert len(msp) == 0
    assert line.dxf.owner == block.dxf.handle


def test_destroyed_modelspace():
    """User destroyed modelspace layout?"""
    doc = ezdxf.new()
    msp = doc.modelspace()
    # You can shoot yourself in the foot if you want:
    msp.block_record.destroy()

    result = doc.audit()
    assert result.has_errors is True, "non-recoverable error expected"
    assert (
        any(error.code == AuditError.MODELSPACE_NOT_FOUND for error in result.errors)
        is True
    )


def test_destroyed_active_paperspace():
    """User destroyed paperspace layouts?"""
    doc = ezdxf.new()
    # You can shoot yourself in the foot if you want:
    for block in doc.blocks:
        if block.is_any_paperspace:
            block.block_record.destroy()

    result = doc.audit()
    assert result.has_errors is True, "non-recoverable error expected"
    assert (
        any(
            error.code == AuditError.ACTIVE_PAPERSPACE_LAYOUT_NOT_FOUND
            for error in result.errors
        )
        is True
    )


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License

import pytest
from ezdxf.math import BoundingBox2d
from ezdxf.fonts import fonts


@pytest.fixture(scope="module", autouse=True)
def check_shape_files_available():
    if not fonts.font_manager.has_font("txt.shx"):
        pytest.skip(reason="required shapefile fonts are not available")


def test_get_font_family_for_shx_files():
    font_face = fonts.font_manager.get_font_face("txt.shx")
    assert font_face.family == "txt"


def test_get_font_family_for_shp_files():
    font_face = fonts.font_manager.get_font_face("txt.shp")
    assert font_face.family == "txt"


def test_get_font_style_for_shx_files():
    font_face = fonts.font_manager.get_font_face("txt.shx")
    assert font_face.style == ".shx"
    # Same weight and width for all shapefile fonts, because this information is not
    # included in the font definition file.
    assert font_face.weight_str == "Normal"
    assert font_face.width_str == "Medium"


def test_best_match_for_shape_files():
    font_face = fonts.find_best_match(family="iso")
    assert font_face.style == ".shx"

    font_face = fonts.find_best_match(family="iso", style=".shp")
    assert font_face.style == ".shp"

    font_face = fonts.find_best_match(family="iso", style=".lff")
    assert font_face.style == ".lff"


def test_width_value_influences_the_best_match_for_shape_files():
    # Note: the width property is used to prioritize font types:
    # 1st .shx; 2nd: .shp; 3rd: .lff
    font_face = fonts.find_best_match(family="iso", width=5)
    assert font_face.filename == "ISO.shx"
    assert font_face.style == ".shx"

    font_face = fonts.find_best_match(family="iso", width=6)
    assert font_face.style == ".shp"

    font_face = fonts.find_best_match(family="iso", width=7)
    assert font_face.style == ".lff"


class TestShapeFileFont:
    @pytest.fixture(scope="class")
    def shx(self):
        return fonts.make_font("txt.shx", 2.5)

    def test_space_width(self, shx):
        assert shx.space_width() == pytest.approx(2.5)

    def test_text_width(self, shx):
        assert shx.text_width("1234") == pytest.approx(9.1666666666)

    def test_text_width_ex(self, shx):
        assert shx.text_width_ex("1234", cap_height=3, width_factor=2) == pytest.approx(
            22
        )

    def test_text_path(self, shx):
        box = BoundingBox2d(shx.text_path("1234").control_vertices())
        assert box.size.x == pytest.approx(9.166666666)
        assert box.size.y == pytest.approx(2.5)

    def test_text_path_ex(self, shx):
        box = BoundingBox2d(
            shx.text_path_ex("1234", cap_height=3, width_factor=2).control_vertices()
        )
        assert box.size.x == pytest.approx(22)
        assert box.size.y == pytest.approx(3)


class TestGlyphCache:
    @pytest.fixture(scope="class")
    def cache(self):
        return fonts.font_manager.get_shapefile_glyph_cache("txt.shx")

    def test_get_glyphs(self, cache):
        glyph = cache.get_shape(ord("A"))
        box = BoundingBox2d(glyph.control_vertices())
        assert box.size.x == 6
        assert box.size.y == 6
        assert glyph is not cache.get_shape(ord("A")), "expected a copy"

    def test_get_advance_width(self, cache):
        glyph = cache.get_shape(ord("A"))
        assert glyph.end.x == 6  # last move_to defines the advance width
        assert cache.get_advance_width(ord("A")) == 6

    def test_get_space_width(self, cache):
        assert cache.space_width == 6

    def test_get_font_measurements(self, cache):
        fm = cache.font_measurements
        assert fm.cap_height == 6
        assert fm.baseline == 0
        assert fm.x_height == 4
        assert fm.descender_height == 2

    def test_path_for_unsupported_glyphs(self, cache):
        box = cache.get_shape(1234)
        assert len(box) == 5

    def test_spaces_are_measured(self, cache):
        assert cache.get_text_length("   ", cache.font_measurements.cap_height) == 18

    def test_spaces_are_not_rendered(self, cache):
        p = cache.get_text_glyph_paths("   ", cache.font_measurements.cap_height)
        assert len(p) == 0


def test_resolve_shx_font_name():
    assert fonts.resolve_shx_font_name("txt", order="s") == "txt.shx"
    assert fonts.resolve_shx_font_name("txt.shx", order="s") == "txt.shx"
    # txt_____.ttf is not available under test
    # prefer LibreCAD fonts:
    assert fonts.resolve_shx_font_name("txt", order="tls") == "standard.lff"
    assert fonts.resolve_shx_font_name("txt.shx", order="tls") == "standard.lff"


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2019 Manfred Moitzi
# License: MIT License
import pytest
import ezdxf
from ezdxf.entities.light import Light
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text

LIGHT = """0
LIGHT
5
0
330
0
100
AcDbEntity
8
0
100
AcDbLight
90
0
1
NAME
70
1
290
1
291
0
40
1.0
72
2
292
0
293
1
73
0
"""


@pytest.fixture
def entity():
    return Light.from_text(LIGHT)


def test_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert "LIGHT" in ENTITY_CLASSES


def test_default_init():
    entity = Light()
    assert entity.dxftype() == "LIGHT"
    assert entity.dxf.handle is None
    assert entity.dxf.owner is None


def test_default_new():
    entity = Light.new(handle="ABBA", owner="0", dxfattribs={})
    assert entity.dxf.version == 0
    assert entity.dxf.name == ""
    assert entity.dxf.type == 1
    assert entity.dxf.status == 1
    assert entity.dxf.plot_glyph == 0
    assert entity.dxf.intensity == 1
    assert entity.dxf.attenuation_type == 2
    assert entity.dxf.use_attenuation_limits == 0
    assert entity.dxf.cast_shadows == 1
    assert entity.dxf.shadow_type == 0


def test_load_from_text(entity):
    assert entity.dxf.version == 0
    assert entity.dxf.name == "NAME"
    assert entity.dxf.type == 1
    assert entity.dxf.status == 1
    assert entity.dxf.plot_glyph == 0
    assert entity.dxf.intensity == 1
    assert entity.dxf.attenuation_type == 2
    assert entity.dxf.use_attenuation_limits == 0
    assert entity.dxf.cast_shadows == 1
    assert entity.dxf.shadow_type == 0


def test_write_dxf():
    entity = Light.from_text(LIGHT)
    result = TagCollector.dxftags(entity)
    expected = basic_tags_from_text(LIGHT)
    assert result == expected


def test_add_light():
    doc = ezdxf.new("R2007")
    msp = doc.modelspace()
    light = msp.new_entity("LIGHT", {"name": "Licht"})
    assert light.dxftype() == "LIGHT"
    assert light.dxf.name == "Licht"



# Purpose: constant values
# Created: 10.03.2011
# Copyright (c) 2011-2018, Manfred Moitzi
# License: MIT License

# --- For Unix ---
#  0:    undefined    Rises an error.
#  1:    ascii        $20-$7F,%%c,%%d,%%p Only
#  2:    iso8859_1    Western Europe(Unix)    $20-$FF,%%c,%%d,%%p
#  3:    iso8859_2    Central Europe(Unix)
#  4:    iso8859_3    Eastern Europe(Unix)
#  5:    iso8859_4    Baltic(Unix)
#  6:    iso8859_5    Cyrillic(Unix)
#  7:    iso8859_6    Arabic(Unix)
#  8:    iso8859_7    Greek(Unix)
#  9:    iso8859_8    Hebrew(Unix)
# 10:    iso8859_9    Turkish(Unix)
#
# --- For DOS/Mac ---
# 11:    dos437    DOS USA
# 12:    dos850    DOS Western Europe
# 13:    dos852    DOS Eastern Europe
# 14:    dos855    IBM Russian
# 15:    dos857    IBM Turkish
# 16:    dos860    DOS Portuguese
# 17:    dos861    DOS Icelandic
# 18:    dos863    DOS Canadian French
# 19:    dos864    DOS Arabic
# 20:    dos865    DOS Norwegian
# 21:    dos869    DOS Greek
# 22:    dos932    DOS Japanese
# 23:    mac-roman    Mac
# 24:    big5    DOS Traditional Chinese
# 25:    ksc5601    Korean Wansung
# 26:    johab    Korean Johab
# 27:    dos866    DOS Russian
# 31:    gb2312    DOS Simplified Chinese
#
# --- For Windows ---
# 28:    ansi_1250    Win Eastern Europe
# 29:    ansi_1251    Win Russian
# 30:    ansi_1252    Win Western Europe(ANSI)
# 32:    ansi_1253    Win Greek
# 33:    ansi_1254    Win Turkish
# 34:    ansi_1255    Win Hebrew
# 35:    ansi_1256    Win Arabic
# 36:    ansi_1257    Win Baltic
# 37:    ansi_874    Win Thai
# 38:    ansi_932    Win Japanese
# 39:    ansi_936    Win Simplified Chinese GB
# 40:    ansi_949    Win Korean Wansung
# 41:    ansi_950    Win Traditional Chinese big5
# 42:    ansi_1361    Win Korean Johab
# 43:    ansi_1200    Unicode (reserved)
# --:    ansi_1258    Win Vietnamese (reserved)

codepage_to_encoding = {
    "874": "cp874",  # Thai,
    "932": "cp932",  # Japanese
    "936": "gbk",  # UnifiedChinese
    "949": "cp949",  # Korean
    "950": "cp950",  # TradChinese
    "1250": "cp1250",  # CentralEurope
    "1251": "cp1251",  # Cyrillic
    "1252": "cp1252",  # WesternEurope
    "1253": "cp1253",  # Greek
    "1254": "cp1254",  # Turkish
    "1255": "cp1255",  # Hebrew
    "1256": "cp1256",  # Arabic
    "1257": "cp1257",  # Baltic
    "1258": "cp1258",  # Vietnam
}

encoding_to_codepage = {
    codec: ansi for ansi, codec in codepage_to_encoding.items()
}


def is_supported_encoding(encoding: str = "cp1252") -> bool:
    return encoding in encoding_to_codepage


def toencoding(dxfcodepage: str) -> str:
    for codepage, encoding in codepage_to_encoding.items():
        if dxfcodepage.endswith(codepage):
            return encoding
    return "cp1252"


def tocodepage(encoding: str) -> str:
    return "ANSI_" + encoding_to_codepage.get(encoding, "1252")



# Copyright (c) 2019 Manfred Moitzi
# License: MIT License
# created 2019-02-15
import pytest

from ezdxf.entities.image import ImageDefReactor
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text

IMAGEDEF_REACTOR = """0
IMAGEDEF_REACTOR
5
0
330
0
100
AcDbRasterImageDefReactor
90
2
330
0
"""


@pytest.fixture
def entity():
    return ImageDefReactor.from_text(IMAGEDEF_REACTOR)


def test_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert "IMAGEDEF_REACTOR" in ENTITY_CLASSES


def test_default_init():
    entity = ImageDefReactor()
    assert entity.dxftype() == "IMAGEDEF_REACTOR"
    assert entity.dxf.handle is None
    assert entity.dxf.owner is None


def test_default_new():
    entity = ImageDefReactor.new(
        handle="ABBA", owner="0", dxfattribs={"image_handle": "FEFE"}
    )
    assert entity.dxf.class_version == 2
    assert entity.dxf.image_handle == "FEFE"


def test_load_from_text(entity):
    assert entity.dxf.class_version == 2
    assert entity.dxf.image_handle == "0"


def test_write_dxf():
    entity = ImageDefReactor.from_text(IMAGEDEF_REACTOR)
    result = TagCollector.dxftags(entity)
    expected = basic_tags_from_text(IMAGEDEF_REACTOR)
    assert result == expected



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import Sequence
from .backend import Backend
from .deps import Vec2, Path
from .properties import Properties


class PolygonBuffer(Backend):
    def __init__(self):
        self.path = Path()
        self.start_new_sub_polygon = False

    def draw_polyline(self, properties: Properties, points: Sequence[Vec2]) -> None:
        if len(points) == 0:
            return
        index = 0
        if self.start_new_sub_polygon:
            self.start_new_sub_polygon = False
            count = len(points)
            while self.path.end.isclose(points[index]):
                index += 1
                if index == count:
                    return
            self.path.move_to(points[index])
        for p in points[index + 1 :]:
            self.path.line_to(p)

    def draw_paths(
        self, properties: Properties, paths: Sequence[Path], filled: bool
    ) -> None:
        if filled:
            raise NotImplementedError()
        for p in paths:
            if len(p) == 0:
                continue
            if self.start_new_sub_polygon:
                self.start_new_sub_polygon = False
                self.path.move_to(p.start)
            self.path.append_path(p)

    def get_paths(self) -> Sequence[Path]:
        return list(self.path.sub_paths())

    def close_path(self):
        if len(self.path):
            self.path.close_sub_path()
            self.start_new_sub_polygon = True

    def reset(self, location: Vec2) -> None:
        self.path = Path(location)
        self.start_new_sub_polygon = False



# Copyright (c) 2014-2022, Manfred Moitzi
# License: MIT License
import pytest

import ezdxf
from ezdxf.sections.acdsdata import AcDsDataSection
from ezdxf import DXFKeyError
from ezdxf.lldxf.tags import internal_tag_compiler, group_tags
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text

HANDLE = "339"


@pytest.fixture
def section():
    entities = group_tags(internal_tag_compiler(ACDSSECTION))
    return AcDsDataSection(None, entities)


def test_loader(section):
    assert "ACDSDATA" == section.name.upper()
    assert len(section.entities) > 0


def test_default_section_does_not_have_records():
    acds = AcDsDataSection(None)
    assert acds.has_records is False


def test_loaded_section_does_not_have_records(section):
    assert section.has_records is True


def test_acds_record(section):
    records = list(section.acdsrecords)
    assert len(records) > 0
    record = records[0]
    assert record.has_section("ASM_Data") is True
    assert record.has_section("AcDbDs::ID") is True
    assert record.has_section("mozman") is False
    with pytest.raises(DXFKeyError):
        _ = record.get_section("mozman")

    asm_data = record.get_section("ASM_Data")
    binary_data = (tag for tag in asm_data if tag.code == 310)
    length = sum(len(tag.value) for tag in binary_data)
    assert asm_data[2].value == length


def test_write_dxf(section):
    result = TagCollector.dxftags(section)
    expected = basic_tags_from_text(ACDSSECTION)
    assert result[:-1] == expected


def test_empty_acds_section_will_not_be_exported():
    doc = ezdxf.new("R2013")
    result = TagCollector.dxftags(doc.acdsdata)
    assert len(result) == 0


def test_get_acis_data(section):
    data = section.get_acis_data(HANDLE)
    assert len(data) == 1088


def test_del_acis_data(section):
    count = len(section.entities)
    section.del_acis_data(HANDLE)
    assert len(section.entities) == count - 1
    assert section.find_acis_record(HANDLE) is None


def test_reset_acis_data(section):
    data = b"0123456789" * 100
    section.set_acis_data(HANDLE, data)
    assert section.get_acis_data(HANDLE) == data


def test_set_new_acis_data(section):
    """Setting data for a non-existing acis-record creates a new record
    automatically.
    """
    handle = "ABBA"
    data = b"0123456789" * 100
    section.set_acis_data(handle, data)
    assert section.get_acis_data(handle) == data


def test_create_explicit_a_new_acis_data(section):
    handle = "ABBA"
    data = b"0123456789" * 100
    section.new_acis_data(handle, data)
    assert section.get_acis_data(handle) == data


def test_write_dxf_for_new_created_acis_record(section):
    data = section.get_acis_data(HANDLE)
    section.del_acis_data(HANDLE)
    section.new_acis_data(HANDLE, data)
    result = TagCollector.dxftags(section)
    expected = basic_tags_from_text(ACDSSECTION)
    assert result[:-1] == expected


ACDSSECTION = """0
SECTION
2
ACDSDATA
70
2
71
6
0
ACDSSCHEMA
90
0
1
AcDb3DSolid_ASM_Data
2
AcDbDs::ID
280
10
91
8
2
ASM_Data
280
15
91
0
101
ACDSRECORD
95
0
90
2
2
AcDbDs::TreatedAsObjectData
280
1
291
1
101
ACDSRECORD
95
0
90
3
2
AcDbDs::Legacy
280
1
291
1
101
ACDSRECORD
1
AcDbDs::ID
90
4
2
AcDs:Indexable
280
1
291
1
101
ACDSRECORD
1
AcDbDs::ID
90
5
2
AcDbDs::HandleAttribute
280
7
282
1
0
ACDSSCHEMA
90
1
1
AcDb_Thumbnail_Schema
2
AcDbDs::ID
280
10
91
8
2
Thumbnail_Data
280
15
91
0
101
ACDSRECORD
95
1
90
2
2
AcDbDs::TreatedAsObjectData
280
1
291
1
101
ACDSRECORD
95
1
90
3
2
AcDbDs::Legacy
280
1
291
1
101
ACDSRECORD
1
AcDbDs::ID
90
4
2
AcDs:Indexable
280
1
291
1
101
ACDSRECORD
1
AcDbDs::ID
90
5
2
AcDbDs::HandleAttribute
280
7
282
1
0
ACDSSCHEMA
90
2
1
AcDbDs::TreatedAsObjectDataSchema
2
AcDbDs::TreatedAsObjectData
280
1
91
0
0
ACDSSCHEMA
90
3
1
AcDbDs::LegacySchema
2
AcDbDs::Legacy
280
1
91
0
0
ACDSSCHEMA
90
4
1
AcDbDs::IndexedPropertySchema
2
AcDs:Indexable
280
1
91
0
0
ACDSSCHEMA
90
5
1
AcDbDs::HandleAttributeSchema
2
AcDbDs::HandleAttribute
280
7
91
1
284
1
0
ACDSRECORD
90
1
2
AcDbDs::ID
280
10
320
339
2
ASM_Data
280
15
94
1088
310
414349532042696E61727946696C652855000000000000020000000C00000007104175746F6465736B204175746F434144071841534D203231392E302E302E3536303020556E6B6E6F776E071853756E204D61792020342031353A34373A3233203230313406000000000000F03F068DEDB5A0F7C6B03E06BBBDD7D9DF7CDB
310
3D0D0961736D6865616465720CFFFFFFFF04FFFFFFFF070C3231392E302E302E35363030110D04626F64790C0200000004FFFFFFFF0CFFFFFFFF0C030000000CFFFFFFFF0CFFFFFFFF110E067265665F76740E036579650D066174747269620CFFFFFFFF04FFFFFFFF0CFFFFFFFF0CFFFFFFFF0C010000000C040000000C05
310
000000110D046C756D700C0600000004FFFFFFFF0CFFFFFFFF0CFFFFFFFF0C070000000C01000000110D0E6579655F726566696E656D656E740CFFFFFFFF04FFFFFFFF070567726964200401000000070374726904010000000704737572660400000000070361646A040000000007046772616404000000000709706F7374
310
636865636B0400000000070463616C6304010000000704636F6E760400000000070473746F6C06000000E001FD414007046E746F6C060000000000003E4007046473696C0600000000000000000708666C61746E6573730600000000000000000707706978617265610600000000000000000704686D617806000000000000
310
0000070667726964617206000000000000000007056D6772696404B80B0000070575677269640400000000070576677269640400000000070A656E645F6669656C6473110D0F7665727465785F74656D706C6174650CFFFFFFFF04FFFFFFFF0403000000040000000004010000000408000000110E067265665F76740E0365
310
79650D066174747269620CFFFFFFFF04FFFFFFFF0CFFFFFFFF0CFFFFFFFF0C030000000C040000000C05000000110D057368656C6C0C0800000004FFFFFFFF0CFFFFFFFF0CFFFFFFFF0CFFFFFFFF0C090000000CFFFFFFFF0C03000000110E067265665F76740E036579650D066174747269620CFFFFFFFF04FFFFFFFF0CFF
310
FFFFFF0CFFFFFFFF0C070000000C040000000C05000000110D04666163650C0A00000004FFFFFFFF0CFFFFFFFF0CFFFFFFFF0CFFFFFFFF0C070000000CFFFFFFFF0C0B0000000B0B110E05666D6573680E036579650D066174747269620CFFFFFFFF04FFFFFFFF0C0C0000000CFFFFFFFF0C09000000110E05746F7275730D
310
07737572666163650CFFFFFFFF04FFFFFFFF0CFFFFFFFF131D7B018BA58BA7C0600EB0424970BC4000000000000000001400000000000000000000000000000000000000000000F03F065087D2E2C5418940066050CEE5F3CA644014000000000000F03F000000000000000000000000000000000B0B0B0B0B110E06726566
310
5F76740E036579650D066174747269620CFFFFFFFF04FFFFFFFF0CFFFFFFFF0C0A0000000C090000000C040000000C05000000110E03456E640E026F660E0341534D0D0464617461
"""



#  Copyright (c) 2021-2024, Manfred Moitzi
#  License: MIT License
from __future__ import annotations

import pathlib
import tempfile
from typing import Callable, Optional, TYPE_CHECKING, Type, Sequence
import abc
import sys
import os
import glob
import signal
import time
import logging
from pathlib import Path

# --------------------------------------------------------------------------------------
# Only imports from the core package here - no add-ons!
#
#   The command `ezdxf -V` must always work!
#
# Imports depending on additional packages like Pillow, Matplotlib, PySide6, ...
# have to be local imports, see 'draw' command as an example.
# --------------------------------------------------------------------------------------
import ezdxf
from ezdxf import recover
from ezdxf.lldxf import const
from ezdxf.lldxf.validator import is_dxf_file, is_binary_dxf_file, dxf_info
from ezdxf.dwginfo import dwg_file_info

if TYPE_CHECKING:
    from ezdxf.entities import DXFGraphic
    from ezdxf.addons.drawing.properties import Properties, LayerProperties

__all__ = ["get", "add_parsers"]

logger = logging.getLogger("ezdxf")


def get(cmd: str) -> Optional[Callable]:
    cls = _commands.get(cmd)
    if cls:
        return cls.run
    return None


def add_parsers(subparsers) -> None:
    for cmd in _commands.values():  # in order of registration
        try:
            cmd.add_parser(subparsers)
        except ImportError:
            logger.info(f"ImportError - '{cmd.NAME}' command not available")


def is_dxf_r12_file(filename: str) -> bool:
    try:
        with open(filename, "rt", errors="ignore") as fp:
            info = dxf_info(fp)
    except IOError:
        return False
    return info.version <= const.DXF12


class Command:
    """abstract base class for launcher commands"""

    NAME = "command"

    @staticmethod
    @abc.abstractmethod
    def add_parser(subparsers) -> None:
        pass

    @staticmethod
    @abc.abstractmethod
    def run(args) -> None:
        pass


_commands: dict[str, Type[Command]] = dict()


def register(cls: Type[Command]):
    """Register a launcher sub-command."""
    _commands[cls.NAME] = cls
    return cls


@register
class Audit(Command):
    """Launcher sub-command: audit"""

    NAME = "audit"

    @staticmethod
    def add_parser(subparsers):
        parser = subparsers.add_parser(Audit.NAME, help="audit and repair DXF files")
        parser.add_argument(
            "files",
            metavar="FILE",
            nargs="+",
            help="audit DXF files",
        )
        parser.add_argument(
            "-s",
            "--save",
            action="store_true",
            help='save recovered files with extension ".rec.dxf" ',
        )
        parser.add_argument(
            "-x",
            "--explore",
            action="store_true",
            help="filters invalid DXF tags, this may load corrupted files but "
            "data loss is very likely",
        )

    @staticmethod
    def run(args):
        def build_outname(name: str) -> str:
            p = Path(name)
            return str(p.parent / (p.stem + ".rec.dxf"))

        def log_fixes(auditor):
            for error in auditor.fixes:
                logger.info("fixed:" + error.message)

        def log_errors(auditor):
            for error in auditor.errors:
                logger.error(error.message)

        def _audit(filename: str) -> None:
            msg = f"auditing file: {filename}"
            print(msg)
            logger.info(msg)
            if args.explore:
                logger.info("explore mode - skipping invalid tags")
            loader = recover.explore if args.explore else recover.readfile
            try:
                doc, auditor = loader(filename)
            except IOError:
                msg = "Not a DXF file or a generic I/O error."
                print(msg)
                logger.error(msg)
                return  # keep on processing additional files
            except const.DXFStructureError as e:
                msg = f"Invalid or corrupted DXF file: {str(e)}"
                print(msg)
                logger.error(msg)
                return  # keep on processing additional files

            if auditor.has_errors:
                auditor.print_error_report()
                log_errors(auditor)
            if auditor.has_fixes:
                auditor.print_fixed_errors()
                log_fixes(auditor)

            if auditor.has_errors is False and auditor.has_fixes is False:
                print("No errors found.")
            else:
                print(
                    f"Found {len(auditor.errors)} errors, "
                    f"applied {len(auditor.fixes)} fixes"
                )

            if args.save:
                outname = build_outname(filename)
                try:
                    doc.saveas(outname)
                except IOError as e:
                    print(f"Can not save recovered file '{outname}':\n{str(e)}")
                else:
                    print(f"Saved recovered file as: '{outname}'")

        for pattern in args.files:
            names = list(glob.glob(pattern))
            if len(names) == 0:
                msg = f"File(s) '{pattern}' not found."
                print(msg)
                logger.error(msg)
                continue
            for filename in names:
                if not os.path.exists(filename):
                    msg = f"File '{filename}' not found."
                    print(msg)
                    logger.error(msg)
                    continue
                if not is_dxf_file(filename):
                    msg = f"File '{filename}' is not a DXF file."
                    print(msg)
                    logger.error(msg)
                    continue
                _audit(filename)


def load_document(filename: str):
    try:
        doc, auditor = recover.readfile(filename)
    except IOError:
        msg = f'Not a DXF file or a generic I/O error: "{filename}"'
        print(msg, file=sys.stderr)
        sys.exit(2)
    except const.DXFStructureError:
        msg = f'Invalid or corrupted DXF file: "{filename}"'
        print(msg, file=sys.stderr)
        sys.exit(3)

    if auditor.has_errors:
        # But is most likely good enough for rendering.
        msg = f"Audit process found {len(auditor.errors)} unrecoverable error(s)."
        print(msg)
        logger.error(msg)
    if auditor.has_fixes:
        msg = f"Audit process fixed {len(auditor.fixes)} error(s)."
        print(msg)
        logger.info(msg)
    return doc, auditor


HELP_LTYPE = (
    "select the line type rendering method, default is approximate. "
    "Approximate uses the closest approximation available to the "
    "backend, the accurate method renders as accurately as possible "
    "but this approach is slower."
)
HELP_LWSCALE = (
    "set custom line weight scaling, default is 0 to disable line " "weights at all"
)


@register
class Draw(Command):
    """Launcher sub-command: draw"""

    NAME = "draw"

    @staticmethod
    def add_parser(subparsers):
        parser = subparsers.add_parser(
            Draw.NAME, help="draw and save DXF as a bitmap or vector image"
        )
        parser.add_argument(
            "file",
            metavar="FILE",
            nargs="?",
            help="DXF file to view or convert",
        )
        parser.add_argument(
            "--backend",
            default="matplotlib",
            choices=["matplotlib", "qt", "mupdf", "custom_svg"],
            help="choose the backend to use for rendering",
        )
        parser.add_argument(
            "--formats",
            action="store_true",
            help="show all supported export formats and exit",
        )
        parser.add_argument(
            "-l",
            "--layout",
            default="Model",
            help='select the layout to draw, default is "Model"',
        )
        parser.add_argument(
            "--background",
            default="DEFAULT",
            choices=[
                "DEFAULT",
                "WHITE",
                "BLACK",
                "PAPERSPACE",
                "MODELSPACE",
                "OFF",
                "CUSTOM",
            ],
            help="choose the background color to use",
        )
        parser.add_argument(
            "--all-layers-visible",
            action="store_true",
            help="draw all layers including the ones marked as invisible",
        )
        parser.add_argument(
            "--all-entities-visible",
            action="store_true",
            help="draw all entities including the ones marked as invisible "
            "(some entities are individually marked as invisible even "
            "if the layer is visible)",
        )
        parser.add_argument(
            "-o",
            "--out",
            required=False,
            type=pathlib.Path,
            help="output filename for export",
        )
        parser.add_argument(
            "--dpi",
            type=int,
            default=300,
            help="target render resolution, default is 300",
        )
        parser.add_argument(
            "-f",
            "--force",
            action="store_true",
            help="overwrite the destination if it already exists",
        )
        parser.add_argument(
            "-v",
            "--verbose",
            action="store_true",
            help="give more output",
        )

    @staticmethod
    def run(args):
        try:
            from ezdxf.addons.drawing import RenderContext, Frontend
            from ezdxf.addons.drawing.config import Configuration, BackgroundPolicy
            from ezdxf.addons.drawing.file_output import (
                open_file,
                MatplotlibFileOutput,
                PyQtFileOutput,
                SvgFileOutput,
                MuPDFFileOutput,
            )
        except ImportError as e:
            print(str(e))
            sys.exit(1)

        if args.backend == "matplotlib":
            try:
                file_output = MatplotlibFileOutput(args.dpi)
            except ImportError as e:
                print(str(e))
                sys.exit(1)
        elif args.backend == "qt":
            try:
                file_output = PyQtFileOutput(args.dpi)
            except ImportError as e:
                print(str(e))
                sys.exit(1)
        elif args.backend == "mupdf":
            try:
                file_output = MuPDFFileOutput(args.dpi)
            except ImportError as e:
                print(str(e))
                sys.exit(1)
        elif args.backend == "custom_svg":
            # has no additional dependencies
            file_output = SvgFileOutput(args.dpi)
        else:
            raise ValueError(args.backend)

        verbose = args.verbose
        if verbose:
            logging.basicConfig(level=logging.INFO)

        if args.formats:
            print(f"formats supported by {args.backend}:")
            for extension, description in file_output.supported_formats():
                print(f"  {extension}: {description}")
            sys.exit(0)

        if args.file:
            filename = args.file
        else:
            print("argument FILE is required")
            sys.exit(1)

        print(f'loading file "{filename}"...')
        doc, _ = load_document(filename)

        try:
            layout = doc.layouts.get(args.layout)
        except KeyError:
            print(
                f'Could not find layout "{args.layout}". '
                f"Valid layouts: {[l.name for l in doc.layouts]}"
            )
            sys.exit(1)

        ctx = RenderContext(doc)
        config = Configuration().with_changes(
            background_policy=BackgroundPolicy[args.background]
        )
        out = file_output.backend()

        if args.all_layers_visible:
            def override_layer_properties(layer_properties: Sequence[LayerProperties]) -> None:
                for properties in layer_properties:
                    properties.is_visible = True

            ctx.set_layer_properties_override(override_layer_properties)

        frontend = Frontend(ctx, out, config=config)

        if args.all_entities_visible:
            def override_entity_properties(entity: DXFGraphic, properties: Properties) -> None:
                properties.is_visible = True
            frontend.push_property_override_function(override_entity_properties)

        t0 = time.perf_counter()
        if verbose:
            print(f"drawing layout '{layout.name}' ...")
        frontend.draw_layout(layout, finalize=True)
        t1 = time.perf_counter()
        if verbose:
            print(f"took {t1-t0:.4f} seconds")

        if args.out is not None:
            if pathlib.Path(args.out).suffix not in {
                f".{ext}" for ext, _ in file_output.supported_formats()
            }:
                print(
                    f'the format of the output path "{args.out}" '
                    f"is not supported by the backend {args.backend}"
                )
                sys.exit(1)

            if args.out.exists() and not args.force:
                print(f'the destination "{args.out}" already exists. Not writing')
                sys.exit(1)
            else:
                print(f'exporting to "{args.out}"...')
                t0 = time.perf_counter()
                file_output.save(args.out)
                t1 = time.perf_counter()
                if verbose:
                    print(f"took {t1 - t0:.4f} seconds")

        else:
            print(f"exporting to temporary file...")
            output_dir = pathlib.Path(tempfile.mkdtemp(prefix="ezdxf_draw"))
            output_path = output_dir / f"output.{file_output.default_format()}"
            file_output.save(output_path)
            print(f'saved to "{output_path}"')
            if verbose:
                print("opening viewer...")
            open_file(output_path)


@register
class View(Command):
    """Launcher sub-command: view"""

    NAME = "view"

    @staticmethod
    def add_parser(subparsers):
        parser = subparsers.add_parser(
            View.NAME, help="view DXF files by the PyQt viewer"
        )
        parser.add_argument(
            "file",
            metavar="FILE",
            nargs="?",
            help="DXF file to view",
        )
        parser.add_argument(
            "-l",
            "--layout",
            default="Model",
            help='select the layout to draw, default is "Model"',
        )
        # disable lineweight at all by default:
        parser.add_argument(
            "--lwscale",
            type=float,
            default=0,
            help=HELP_LWSCALE,
        )

    @staticmethod
    def run(args):
        # Import on demand for a quicker startup:
        try:
            from ezdxf.addons.xqt import QtWidgets
        except ImportError as e:
            print(str(e))
            sys.exit(1)
        from ezdxf.addons.drawing.qtviewer import CADViewer
        from ezdxf.addons.drawing.config import Configuration

        config = Configuration(
            lineweight_scaling=args.lwscale,
        )

        signal.signal(signal.SIGINT, signal.SIG_DFL)  # handle Ctrl+C properly
        app = QtWidgets.QApplication(sys.argv)
        app.setStyle("Fusion")
        set_app_icon(app)
        viewer = CADViewer.from_config(config)
        filename = args.file
        if filename:
            doc, auditor = load_document(filename)
            viewer.set_document(
                doc,
                auditor,
                layout=args.layout,
            )
        sys.exit(app.exec())


@register
class Browse(Command):
    """Launcher sub-command: browse"""

    NAME = "browse"

    @staticmethod
    def add_parser(subparsers):
        parser = subparsers.add_parser(Browse.NAME, help="browse DXF file structure")
        parser.add_argument(
            "file",
            metavar="FILE",
            nargs="?",
            help="DXF file to browse",
        )
        parser.add_argument(
            "-l", "--line", type=int, required=False, help="go to line number"
        )
        parser.add_argument(
            "-g",
            "--handle",
            required=False,
            help="go to entity by HANDLE, HANDLE has to be a hex value without "
            "any prefix like 'fefe'",
        )

    @staticmethod
    def run(args):
        try:
            from ezdxf.addons.xqt import QtWidgets
        except ImportError as e:
            print(str(e))
            sys.exit(1)
        from ezdxf.addons import browser

        signal.signal(signal.SIGINT, signal.SIG_DFL)  # handle Ctrl+C properly
        app = QtWidgets.QApplication(sys.argv)
        app.setStyle("Fusion")
        set_app_icon(app)
        main_window = browser.DXFStructureBrowser(
            args.file,
            line=args.line,
            handle=args.handle,
            resource_path=resources_path(),
        )
        main_window.show()
        sys.exit(app.exec())


@register
class BrowseAcisData(Command):
    """Launcher sub-command: browse-acis"""

    NAME = "browse-acis"

    @staticmethod
    def add_parser(subparsers):
        parser = subparsers.add_parser(
            BrowseAcisData.NAME, help="browse ACIS structures in DXF files"
        )
        parser.add_argument(
            "file",
            metavar="FILE",
            nargs="?",
            help="DXF file to browse",
        )
        parser.add_argument(
            "-g",
            "--handle",
            required=False,
            help="go to entity by HANDLE, HANDLE has to be a hex value without "
            "any prefix like 'fefe'",
        )

    @staticmethod
    def run(args):
        try:
            from ezdxf.addons.xqt import QtWidgets
        except ImportError as e:
            print(str(e))
            sys.exit(1)
        from ezdxf.addons.acisbrowser.browser import AcisStructureBrowser

        signal.signal(signal.SIGINT, signal.SIG_DFL)  # handle Ctrl+C properly
        app = QtWidgets.QApplication(sys.argv)
        app.setStyle("Fusion")
        set_app_icon(app)
        main_window = AcisStructureBrowser(
            args.file,
            handle=args.handle,
        )
        main_window.show()
        sys.exit(app.exec())


@register
class Strip(Command):
    """Launcher sub-command: strip"""

    NAME = "strip"

    @staticmethod
    def add_parser(subparsers):
        parser = subparsers.add_parser(Strip.NAME, help="strip comments from DXF files")
        parser.add_argument(
            "file",
            metavar="FILE",
            nargs="+",
            help='DXF file to process, wildcards "*" and "?" are supported',
        )
        parser.add_argument(
            "-b",
            "--backup",
            action="store_true",
            required=False,
            help='make a backup copy with extension ".bak" from the '
            "DXF file, overwrites existing backup files",
        )
        parser.add_argument(
            "-t",
            "--thumbnail",
            action="store_true",
            required=False,
            help="strip THUMBNAILIMAGE section",
        )
        parser.add_argument(
            "--handles",
            action="store_true",
            required=False,
            help="remove handles from DXF R12 or older files",
        )
        parser.add_argument(
            "-v",
            "--verbose",
            action="store_true",
            required=False,
            help="give more output",
        )

    @staticmethod
    def run(args):
        from ezdxf.tools.strip import strip

        for pattern in args.file:
            for filename in glob.glob(pattern):
                codes = [999]
                if args.handles:
                    if is_dxf_r12_file(filename):
                        codes.extend([5, 105])
                    else:
                        print(
                            f"Cannot remove handles from DXF R13 or later: {filename}"
                        )
                strip(
                    filename,
                    backup=args.backup,
                    thumbnail=args.thumbnail,
                    verbose=args.verbose,
                    codes=codes,
                )


@register
class Config(Command):
    """Launcher sub-command: config"""

    NAME = "config"

    @staticmethod
    def add_parser(subparsers):
        parser = subparsers.add_parser(Config.NAME, help="manage config files")
        parser.add_argument(
            "-p",
            "--print",
            action="store_true",
            help="print configuration",
        )
        parser.add_argument(
            "-w",
            "--write",
            metavar="FILE",
            help="write configuration",
        )
        parser.add_argument(
            "--home",
            action="store_true",
            help="create config file 'ezdxf.ini' in the user home directory "
            "'~/.config/ezdxf', $XDG_CONFIG_HOME is supported if set",
        )
        parser.add_argument(
            "--reset",
            action="store_true",
            help="factory reset, delete default config files 'ezdxf.ini'",
        )

    @staticmethod
    def run(args):
        from ezdxf import options

        action = False
        if args.reset:
            options.reset()
            options.delete_default_config_files()
            action = True
        if args.home:
            options.write_home_config()
            action = True
        if args.write:
            action = True
            filepath = Path(args.write).expanduser()
            try:
                options.write_file(str(filepath))
                print(f"configuration written to: {filepath}")
            except IOError as e:
                print(str(e))
        if args.print or action is False:
            options.print()


def load_every_document(filename: str):
    def io_error() -> str:
        msg = f'Not a DXF file or a generic I/O error: "{filename}"'
        print(msg, file=sys.stderr)
        return msg

    def structure_error() -> str:
        msg = f'Invalid or corrupted DXF file: "{filename}"'
        print(msg, file=sys.stderr)
        return msg

    binary_fmt = False
    if is_binary_dxf_file(filename):
        try:
            doc = ezdxf.readfile(filename)
        except IOError:
            raise const.DXFLoadError(io_error())
        except const.DXFStructureError:
            raise const.DXFLoadError(structure_error())
        auditor = doc.audit()
        binary_fmt = True
    else:
        try:
            doc, auditor = recover.readfile(filename)
        except IOError:
            raise const.DXFLoadError(io_error())
        except const.DXFStructureError:
            dwginfo = dwg_file_info(filename)
            if dwginfo.version != "invalid":
                print(
                    f"This is a DWG file!!!\n"
                    f'Filename: "{filename}"\n'
                    f"Format: DWG\n"
                    f"Release: {dwginfo.release}\n"
                    f"DWG Version: {dwginfo.version}\n"
                )
                raise const.DXFLoadError()
            raise const.DXFLoadError(structure_error())
    return doc, auditor, binary_fmt


@register
class Info(Command):
    """Launcher sub-command: info"""

    NAME = "info"

    @staticmethod
    def add_parser(subparsers):
        parser = subparsers.add_parser(
            Info.NAME,
            help="show information and optional stats of DXF files as "
            "loaded by ezdxf, this may not represent the original "
            "content of the file, use the browse command to "
            "see the original content",
        )
        parser.add_argument(
            "file",
            metavar="FILE",
            nargs="+",
            help='DXF file to process, wildcards "*" and "?" are supported',
        )
        parser.add_argument(
            "-v",
            "--verbose",
            action="store_true",
            required=False,
            help="give more output",
        )
        parser.add_argument(
            "-s",
            "--stats",
            action="store_true",
            required=False,
            help="show content stats",
        )

    @staticmethod
    def run(args):
        from ezdxf.document import info

        def process(fn: str):
            try:
                doc, auditor, binary_fmt = load_every_document(fn)
            except const.DXFLoadError:
                pass
            else:
                fmt = "Binary" if binary_fmt else "ASCII"
                print(
                    "\n".join(
                        info(
                            doc,
                            verbose=args.verbose,
                            content=args.stats,
                            fmt=fmt,
                        )
                    )
                )
                if auditor.has_fixes:
                    print(f"Audit process fixed {len(auditor.fixes)} error(s).")
                if auditor.has_errors:
                    print(
                        f"Audit process found {len(auditor.errors)} unrecoverable error(s)."
                    )
                print()

        for pattern in args.file:
            file_count = 0
            for filename in glob.glob(pattern):
                if os.path.isdir(filename):
                    dir_pattern = os.path.join(filename, "*.dxf")
                    for filename2 in glob.glob(dir_pattern):
                        process(filename2)
                        file_count += 1
                else:
                    process(filename)
                    file_count += 1

            if file_count == 0:
                sys.stderr.write(f'No matching files for pattern: "{pattern}"\n')


@register
class HPGL(Command):
    """Launcher sub-command: hpgl"""

    NAME = "hpgl"

    @staticmethod
    def add_parser(subparsers):
        parser = subparsers.add_parser(
            HPGL.NAME, help=f"view and/or convert HPGL/2 plot files to various formats"
        )
        parser.add_argument(
            "file",
            metavar="FILE",
            nargs="?",
            default="",
            help=f"view and/or convert HPGL/2 plot files, wildcards (*, ?) supported in command line mode",
        )
        parser.add_argument(
            "-e",
            "--export",
            metavar="FORMAT",
            required=False,
            help=f"convert HPGL/2 plot file to SVG, PDF or DXF from the command line (no gui)",
        )
        parser.add_argument(
            "-r",
            "--rotate",
            type=int,
            choices=(0, 90, 180, 270),
            default=0,
            required=False,
            help="rotate page about 90, 180 or 270 degrees (no gui)",
        )
        parser.add_argument(
            "-x",
            "--mirror_x",
            action="store_true",
            required=False,
            help="mirror page in x-axis direction, (no gui)",
        )
        parser.add_argument(
            "-y",
            "--mirror_y",
            action="store_true",
            required=False,
            help="mirror page in y-axis direction, (no gui)",
        )
        parser.add_argument(
            "-m",
            "--merge_control",
            type=int,
            required=False,
            default=2,
            choices=(0, 1, 2),
            help="provides control over the order of filled polygons, 0=off (print order), "
            "1=luminance (order by luminance), 2=auto (default)",
        )
        parser.add_argument(
            "-f",
            "--force",
            action="store_true",
            required=False,
            help="inserts the mandatory 'enter HPGL/2 mode' escape sequence into the data "
            "stream; use this flag when no HPGL/2 data was found and you are sure the "
            "file is a HPGL/2 plot file",
        )
        parser.add_argument(
            "--aci",
            action="store_true",
            required=False,
            help="use pen numbers as ACI colors and assign colors by layer (DXF only)",
        )
        parser.epilog = (
            "Note that plot files are intended to be plotted on white paper."
        )
        parser.add_argument(
            "--dpi",
            type=int,
            required=False,
            default=96,
            help="pixel density in dots per inch (PNG only)",
        )
        parser.epilog = (
            "Note that plot files are intended to be plotted on white paper."
        )

    @staticmethod
    def run(args):
        if args.export:
            if os.path.exists(args.file):
                filenames = [args.file]
            else:
                filenames = glob.glob(args.file)
            for filename in filenames:
                export_hpgl2(Path(filename), args)
        else:
            launch_hpgl2_viewer(args.file, args.force)


def export_hpgl2(filepath: Path, args) -> None:
    from ezdxf.addons.hpgl2 import api as hpgl2
    from ezdxf.addons.drawing.dxf import ColorMode

    fmt = args.export.upper()
    start_msg = f"converting HPGL/2 plot file '{filepath.name}' to {fmt}"
    try:
        data = filepath.read_bytes()
    except IOError as e:
        print(str(e), file=sys.stderr)
        return
    if args.force:
        data = b"%1B" + data
    export_path = filepath.with_suffix(f".{fmt.lower()}")
    if fmt == "DXF":
        print(start_msg)
        color_mode = ColorMode.ACI if args.aci else ColorMode.RGB
        doc = hpgl2.to_dxf(
            data,
            rotation=args.rotate,
            mirror_x=args.mirror_x,
            mirror_y=args.mirror_y,
            color_mode=color_mode,
            merge_control=args.merge_control,
        )
        try:
            doc.saveas(export_path)
        except IOError as e:
            print(str(e), file=sys.stderr)

    elif fmt == "SVG":
        print(start_msg)
        svg_string = hpgl2.to_svg(
            data,
            rotation=args.rotate,
            mirror_x=args.mirror_x,
            mirror_y=args.mirror_y,
            merge_control=args.merge_control,
        )
        try:
            export_path.write_text(svg_string)
        except IOError as e:
            print(str(e), file=sys.stderr)
    elif fmt == "PDF":
        print(start_msg)
        pdf_bytes = hpgl2.to_pdf(
            data,
            rotation=args.rotate,
            mirror_x=args.mirror_x,
            mirror_y=args.mirror_y,
            merge_control=args.merge_control,
        )
        try:
            export_path.write_bytes(pdf_bytes)
        except IOError as e:
            print(str(e), file=sys.stderr)
    elif fmt == "PNG":
        print(start_msg)
        png_bytes = hpgl2.to_pixmap(
            data,
            rotation=args.rotate,
            mirror_x=args.mirror_x,
            mirror_y=args.mirror_y,
            merge_control=args.merge_control,
            fmt="png",
            dpi=args.dpi,
        )
        try:
            export_path.write_bytes(png_bytes)
        except IOError as e:
            print(str(e), file=sys.stderr)
    else:
        print(f"invalid export format: {fmt}")
        exit(1)
    print(f"file '{export_path.name}' successfully written")


def launch_hpgl2_viewer(filename: str, force: bool) -> None:
    try:
        from ezdxf.addons.xqt import QtWidgets
    except ImportError as e:
        print(str(e))
        exit(1)
    from ezdxf.addons.hpgl2.viewer import HPGL2Viewer

    signal.signal(signal.SIGINT, signal.SIG_DFL)  # handle Ctrl+C properly
    app = QtWidgets.QApplication(sys.argv)
    app.setStyle("Fusion")
    set_app_icon(app)
    viewer = HPGL2Viewer()
    viewer.show()
    if filename and os.path.exists(filename):
        viewer.load_plot_file(filename, force=force)
    sys.exit(app.exec())


def set_app_icon(app):
    from ezdxf.addons.xqt import QtGui, QtCore

    app_icon = QtGui.QIcon()
    p = resources_path()
    app_icon.addFile(str(p / "16x16.png"), QtCore.QSize(16, 16))
    app_icon.addFile(str(p / "24x24.png"), QtCore.QSize(24, 24))
    app_icon.addFile(str(p / "32x32.png"), QtCore.QSize(32, 32))
    app_icon.addFile(str(p / "48x48.png"), QtCore.QSize(48, 48))
    app_icon.addFile(str(p / "64x64.png"), QtCore.QSize(64, 64))
    app_icon.addFile(str(p / "256x256.png"), QtCore.QSize(256, 256))
    app.setWindowIcon(app_icon)


def resources_path():
    from pathlib import Path

    return Path(__file__).parent / "resources"



#  Copyright (c) 2021-2024, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
import pytest
from ezdxf.math import Vec2, BoundingBox2d
from ezdxf.math.clipping import (
    ConvexClippingPolygon2d,
    ClippingRect2d,
    Clipping,
)
from ezdxf.render.forms import circle


class TestClipSingleLineAtConvexBoundary:
    @pytest.fixture(scope="class", params=["polygon", "rect"])
    def clipper(self, request):
        if request.param == "polygon":
            return ConvexClippingPolygon2d(Vec2.list([(0, 0), (2, 0), (2, 2), (0, 2)]))
        else:
            return ClippingRect2d(Vec2(0, 0), Vec2(2, 2))

    def test_no_clipping(self, clipper):
        assert len(clipper.clip_line(Vec2(-1, 3), Vec2(3, 3))) == 0  # above
        assert len(clipper.clip_line(Vec2(-1, -1), Vec2(3, -1))) == 0  # below
        assert len(clipper.clip_line(Vec2(-1, 0), Vec2(-1, 2))) == 0  # left
        assert len(clipper.clip_line(Vec2(3, 0), Vec2(3, 2))) == 0  # right

    def test_regular_clip_inside_outside(self, clipper: Clipping):
        s, e = clipper.clip_line(Vec2(1, 1), Vec2(3, 1))[0]
        assert s.isclose((1, 1))
        assert e.isclose((2, 1))

    def test_regular_clip_outside_inside(self, clipper: Clipping):
        s, e = clipper.clip_line(Vec2(3, 1), Vec2(1, 1))[0]
        assert s.isclose((2, 1))
        assert e.isclose((1, 1))

    def test_crossing_horizontal_left_to_right(self, clipper: Clipping):
        s, e = clipper.clip_line(Vec2(-1, 1), Vec2(3, 1))[0]
        assert s.isclose((0, 1))
        assert e.isclose((2, 1))

    def test_crossing_horizontal_right_to_left(self, clipper: Clipping):
        s, e = clipper.clip_line(Vec2(3, 1), Vec2(-1, 1))[0]
        assert s.isclose((2, 1))
        assert e.isclose((0, 1))

    def test_crossing_vertical(self, clipper: Clipping):
        s, e = clipper.clip_line(Vec2(1, -1), Vec2(1, 3))[0]
        assert s.isclose((1, 0))
        assert e.isclose((1, 2))

    def test_crossing_diagonal(self, clipper: Clipping):
        s, e = clipper.clip_line(Vec2(-1, 0), Vec2(2, 3))[0]
        assert s.isclose((0, 1))
        assert e.isclose((1, 2))

    def test_crossing_diagonal_edge_to_edge(self, clipper: Clipping):
        s, e = clipper.clip_line(Vec2(0, 0), Vec2(2, 2))[0]
        assert s.isclose((0, 0))
        assert e.isclose((2, 2))

    @pytest.mark.parametrize("y", [0, 2], ids=["bottom", "top"])
    def test_colinear_horizontal_edge(self, clipper: Clipping, y: int):
        s, e = clipper.clip_line(Vec2(-1, y), Vec2(3, y))[0]
        assert s.isclose((0, y))
        assert e.isclose((2, y))

    @pytest.mark.parametrize("x", [0, 2], ids=["left", "right"])
    def test_colinear_vertical_edge(self, clipper: Clipping, x: int):
        s, e = clipper.clip_line(Vec2(x, -1), Vec2(x, 3))[0]
        assert s.isclose((x, 0))
        assert e.isclose((x, 2))


class TestClipPolylineAtConvexBoundary:
    @pytest.fixture(scope="class", params=["polygon", "rect"])
    def clipper(self, request):
        if request.param == "polygon":
            return ConvexClippingPolygon2d(Vec2.list([(0, 0), (8, 0), (8, 2), (0, 2)]))
        else:
            return ClippingRect2d(Vec2(0, 0), Vec2(8, 2))

    def test_crossing_zigzag(self, clipper: Clipping):
        p0, p1 = clipper.clip_polyline(Vec2.list([(0, -1), (4, 3), (8, -1)]))
        assert p0[0].isclose((1, 0))
        assert p0[1].isclose((3, 2))
        assert p1[0].isclose((5, 2))
        assert p1[1].isclose((7, 0))

    def test_closed_rectangle(self):
        # 9  ...+--f...
        # 8  .d-x==x-c.
        # 7  .|.|..|.|.
        # 6  .|.|..|.|.
        # 5  .|.|..|.|.
        # 4  .|.|..|.|.
        # 3  .|.|..|.|.
        # 2  .|.|..|.|.
        # 1  .a-x==x-b.
        # 0  ...e--+...
        #    0123456789
        #                  a       b       c       d       a
        rect = Vec2.list([(1, 1), (8, 1), (8, 8), (1, 8), (1, 1)])
        #                             e           f
        clipper = ClippingRect2d(Vec2(3, 0), Vec2(6, 9))
        result = clipper.clip_polyline(rect)
        assert len(result) == 2

        bbox = BoundingBox2d(result[0])
        assert bbox.extmin.isclose((3, 1))
        assert bbox.extmax.isclose((6, 1))

        bbox = BoundingBox2d(result[1])
        assert bbox.extmin.isclose((3, 8))
        assert bbox.extmax.isclose((6, 8))


class TestClipPolygonAtConvexBoundary:
    @pytest.fixture(scope="class", params=["polygon", "rect"])
    def clipper(self, request):
        if request.param == "polygon":
            return ConvexClippingPolygon2d(
                Vec2.list([(-1, -1), (1, -1), (1, 1), (-1, 1)])
            )
        else:
            return ClippingRect2d(Vec2(-1, -1), Vec2(1, 1))

    @pytest.fixture
    def rect(self):
        return Vec2.list([(-1, -1), (1, -1), (1, 1), (-1, 1)])

    @pytest.fixture
    def overlapping(self):  # overlapping
        return Vec2.list([(0, 0), (2, 0), (2, 2), (0, 2)])

    @pytest.fixture
    def inside(self):  # complete inside
        return Vec2.list([(0, 0), (0.5, 0), (0.5, 0.5), (0, 0.5)])

    @pytest.fixture
    def outside(self):  # complete outside
        return Vec2.list([(1, 1), (2, 1), (2, 2), (1, 2)])

    def test_subject_do_overlap_clipping_rect(
        self, clipper: Clipping, overlapping: list[Vec2]
    ):
        result = clipper.clip_polygon(overlapping)[0]
        assert len(result) == 4
        assert Vec2(0, 0) in result
        assert Vec2(1, 0) in result
        assert Vec2(1, 1) in result
        assert Vec2(0, 1) in result

    def test_subject_is_inside_rect(self, clipper: Clipping, inside: list[Vec2]):
        result = clipper.clip_polygon(inside)[0]
        assert len(result) == 4
        for v in inside:
            assert any(r.isclose(v) for r in result) is True

    def test_clockwise_oriented_clipping_rect(
        self, rect: list[Vec2], inside: list[Vec2]
    ):
        rect.reverse()
        clipper = ConvexClippingPolygon2d(rect)
        result = clipper.clip_polygon(inside)[0]
        assert len(result) == 4
        for v in inside:
            assert any(r.isclose(v) for r in result) is True

    def test_subject_is_outside_rect(self, clipper: Clipping, outside: list[Vec2]):
        result = clipper.clip_polygon(outside)[0]
        assert len(result) == 0

    def test_circle_outside_rect(self, clipper: Clipping, rect: list[Vec2]):
        c = Vec2.list(circle(16, 3))
        result = clipper.clip_polygon(c)[0]
        assert len(result) == 4
        for v in rect:
            assert any(r.isclose(v) for r in result) is True

    def test_circle_inside_rect(self, clipper: Clipping):
        c = Vec2.list(circle(16, 0.7))
        result = clipper.clip_polygon(c)[0]
        assert len(result) == 16
        for v in c:
            assert any(r.isclose(v) for r in result) is True

    def test_rect_outside_circle(self, rect: list[Vec2]):
        c = Vec2.list(circle(16, 0.7))
        clipper = ConvexClippingPolygon2d(c)
        result = clipper.clip_polygon(rect)[0]
        assert len(result) == 16
        for v in c:
            assert any(r.isclose(v) for r in result) is True

    def test_rect_inside_circle(self, rect: list[Vec2]):
        c = Vec2.list(circle(16, 3))
        clipper = ConvexClippingPolygon2d(c)
        result = clipper.clip_polygon(rect)[0]
        assert len(result) == 4
        for v in rect:
            assert any(r.isclose(v) for r in result) is True


def test_imprecisions_in_edge_intersection():
    clipper = ConvexClippingPolygon2d(
        [
            Vec2(8.000000000000455, 9.000000000000165),
            Vec2(15.000000000000887, 9.000000000000165),
            Vec2(15.000000000000887, 11.000000000000834),
            Vec2(8.000000000000455, 11.000000000000834),
        ]
    )
    polygon = [
        Vec2(8.000000000000435, 9.000000000000169),
        Vec2(15.000000000000874, 9.000000000000169),
        Vec2(15.000000000000895, 11.00000000000084),
        Vec2(8.000000000000464, 11.00000000000084),
        Vec2(8.000000000000435, 9.000000000000377),
        Vec2(8.000000000000435, 9.000000000000169),
    ]
    result = clipper.clip_polygon(polygon)
    assert len(result) > 0


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2020, Manfred Moitzi
#  License: MIT License
import os
import pytest
import ezdxf
from ezdxf import recover

BASEDIR = os.path.dirname(__file__)
DATADIR = "data"


@pytest.fixture(params=["AC1003_LINE_Example.dxf"])
def filename(request):
    filename = os.path.join(BASEDIR, DATADIR, request.param)
    if not os.path.exists(filename):
        pytest.skip("File {} not found.".format(filename))
    return filename


def test_coordinate_order_problem(filename):
    try:
        doc, auditor = recover.readfile(filename)
    except ezdxf.DXFError as e:
        pytest.fail(str(e))
    else:
        msp = doc.modelspace()
        lines = msp.query("LINE")
        assert lines[0].dxf.start == (1.5, 0, 0)



# Copyright (c) 2019 Manfred Moitzi
# License: MIT License
import pytest

from ezdxf.lldxf.validator import entity_structure_validator
from ezdxf.lldxf.tagger import internal_tag_compiler
from ezdxf import DXFAppDataError, DXFXDataError


def compile(s):
    return list(internal_tag_compiler(s))


def test_valid_tags():
    tags = list(entity_structure_validator(compile(VALID_ENTITY)))
    assert len(tags) == 6


# The structure validator does not know anything about entities, except its basic tag structure
VALID_ENTITY = """0
LINE
102
{APP
40
1
102
}
1001
TEST
1000
STRING
"""


def test_invalid_app_data_without_closing_tag():
    with pytest.raises(DXFAppDataError):
        list(
            entity_structure_validator(compile(INVALID_APPDATA_NO_CLOSING_TAG))
        )


INVALID_APPDATA_NO_CLOSING_TAG = """0
LINE
102
{APP
40
1
1001
TEST
1000
STRING
"""


def test_invalid_app_data_without_opening_tag():
    with pytest.raises(DXFAppDataError):
        list(
            entity_structure_validator(compile(INVALID_APPDATA_NO_OPENING_TAG))
        )


INVALID_APPDATA_NO_OPENING_TAG = """0
LINE
102
}
40
1
1001
TEST
1000
STRING
"""


def test_invalid_app_data_structure_tag():
    with pytest.raises(DXFAppDataError):
        list(entity_structure_validator(compile(INVALID_APPDATA_STRUCTURE_TAG)))


INVALID_APPDATA_STRUCTURE_TAG = """0
LINE
102
INVALID_APPDATA_STRUCTURE_TAG
40
1
1001
TEST
1000
STRING
"""


def test_xrecord_with_group_code_102():
    tags = list(
        entity_structure_validator(compile(XRECORD_WITH_GROUP_CODE_102))
    )
    assert tags[0] == (0, "XRECORD")


XRECORD_WITH_GROUP_CODE_102 = """0
XRECORD
5
4E6F
102
{ACAD_REACTORS
330
4E50
102
}
330
4E50
100
AcDbXrecord
280
1
1
dik
40
20.0
1
bt.var
102
{
102
{
1
<!I@M:DOT!>
1
kbez
102
}
102
}
"""


def test_xrecord_with_long_closing_tag():
    tags = list(
        entity_structure_validator(compile(XRECORD_APP_DATA_LONG_CLOSING_TAG))
    )
    assert tags[0] == (0, "XRECORD")


XRECORD_APP_DATA_LONG_CLOSING_TAG = """  0
XRECORD
5
2F9
102
{ACAD_REACTORS
330
2FF
102
}
330
2FF
100
AcDbXrecord
280
1
1
AcDb_Thumbnail_Schema
102
{ATTRRECORD
341
2FA
2
AcDbDs::TreatedAsObjectData
280
1
291
1
102
ATTRRECORD}
102
{ATTRRECORD
341
2FB
2
AcDbDs::Legacy
280
1
291
1
102
ATTRRECORD}
2
AcDbDs::ID
280
10
91
8
102
{ATTRRECORD
341
2FC
2
AcDs:Indexable
280
1
291
1
102
ATTRRECORD}
102
{ATTRRECORD
341
2FD
2
AcDbDs::HandleAttribute
280
7
282
1
102
ATTRRECORD}
2
Thumbnail_Data
280
15
91
0
"""


def test_invalid_xdata():
    with pytest.raises(DXFXDataError):
        list(entity_structure_validator(compile(INVALID_XDATA_STRUCTURE_TAG)))


INVALID_XDATA_STRUCTURE_TAG = """0
LINE
40
1
1001
TEST
1000
STRING
1
NO GROUP CODE < 1000 in XDATA SECTION
"""


def test_unbalanced_xdata_list_1():
    with pytest.raises(DXFXDataError):
        list(entity_structure_validator(compile(UNBALANCED_XDATA_LIST_1)))


UNBALANCED_XDATA_LIST_1 = """0
LINE
40
1
1001
TEST
1000
STRING
1002
{
1002
{
1002
}
"""


def test_unbalanced_xdata_list_2():
    with pytest.raises(DXFXDataError):
        list(entity_structure_validator(compile(UNBALANCED_XDATA_LIST_2)))


UNBALANCED_XDATA_LIST_2 = """0
LINE
40
1
1001
TEST
1000
STRING
1002
{
1002
{
1002
}
"""


def test_invalid_xdata_list_nesting():
    with pytest.raises(DXFXDataError):
        list(entity_structure_validator(compile(INVALID_XDATA_LIST_NESTING)))


INVALID_XDATA_LIST_NESTING = """0
LINE
40
1
1001
TEST
1000
STRING
1002
{
1002
}
1002
}
1002
{
"""

BLOCKBASEPOINTPARAMETER_CVIL_3D_2018 = """0
BLOCKBASEPOINTPARAMETER
5
4C25
330
4C23
100
AcDbEvalExpr
90
1
98
33
99
4
100
AcDbBlockElement
300
Base Point
98
33
99
4
1071
0
100
AcDbBlockParameter
280
1
281
0
100
AcDbBlock1PtParameter
1010
-3.108080399920343
1020
-0.9562299080084814
1030
0.0
93
0
170
0
171
0
100
AcDbBlockBasepointParameter
1011
0.0
1021
0.0
1031
0.0
1012
0.0
1022
0.0
1032
0.0
"""


def test_extended_group_code_before_XDATA():
    tags = list(
        entity_structure_validator(
            compile(BLOCKBASEPOINTPARAMETER_CVIL_3D_2018)
        )
    )
    assert tags[0] == (0, "BLOCKBASEPOINTPARAMETER")


ENTITY_W_EMBEDDED_OBJECT = """0
LINE
102
{APP
40
1
102
}
1001
TEST
1000
STRING
101
Embedded Object
1
Text
"""


def test_embedded_object_follows_XDATA():
    tags = list(entity_structure_validator(compile(ENTITY_W_EMBEDDED_OBJECT)))
    assert len(tags) == 8
    assert tags[-1] == (1, "Text")
